# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2017 Sophist-UK
# Copyright (C) 2018, 2020 Wieland Hoffmann
# Copyright (C) 2018-2021 Laurent Monin
# Copyright (C) 2018-2021, 2023 Philipp Wolfer
# Copyright (C) 2020 dukeyin
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from test.picardtestcase import (
    PicardTestCase,
    create_fake_png,
    load_test_json,
)
from test.test_coverart_image import create_image

from picard.acoustid.json_helpers import (
    parse_recording as acoustid_parse_recording,
)
from picard.cluster import Cluster
from picard.coverart.image import CoverArtImage
from picard.file import File
from picard.mbjson import (
    release_to_metadata,
    track_to_metadata,
)
from picard.metadata import (
    MULTI_VALUED_JOINER,
    Metadata,
    MultiMetadataProxy,
    trackcount_score,
    weights_from_preferred_countries,
    weights_from_preferred_formats,
    weights_from_release_type_scores,
)
from picard.track import Track
from picard.util.imagelist import ImageList
from picard.util.tags import PRESERVED_TAGS


settings = {
    'write_id3v23': False,
    'id3v23_join_with': '/',
    'preferred_release_countries': [],
    'preferred_release_formats': [],
    'standardize_artists': False,
    'standardize_instruments': False,
    'translate_artist_names': False,
    'release_ars': True,
    'release_type_scores': [
        ('Album', 1.0),
        ('Other', 1.0),
    ],
}


class CommonTests:

    class CommonMetadataTestCase(PicardTestCase):

        original = None
        tags = []

        def setUp(self):
            super().setUp()
            self.set_config_values(settings)
            self.metadata = self.get_metadata_object()
            self.metadata.length = 242
            self.metadata["single1"] = "single1-value"
            self.metadata.add_unique("single2", "single2-value")
            self.metadata.add_unique("single2", "single2-value")
            self.multi1 = ["multi1-value1", "multi1-value1"]
            self.metadata.add("multi1", self.multi1[0])
            self.metadata.add("multi1", self.multi1[1])
            self.multi2 = ["multi2-value1", "multi2-value2"]
            self.metadata["multi2"] = self.multi2
            self.multi3 = ["multi3-value1", "multi3-value2"]
            self.metadata.set("multi3", self.multi3)
            self.metadata["~hidden"] = "hidden-value"

            self.metadata_d1 = Metadata({'a': 'b', 'c': 2, 'd': ['x', 'y'], 'x': ''})
            self.metadata_d2 = Metadata({'a': 'b', 'c': 2, 'd': ['x', 'y'], 'x': 'z'})
            self.metadata_d3 = Metadata({'c': 3, 'd': ['u', 'w'], 'x': 'p'})

        @staticmethod
        def get_metadata_object():
            pass

        def tearDown(self):
            pass

        def test_metadata_setitem(self):
            self.assertEqual(["single1-value"], self.metadata.getraw("single1"))
            self.assertEqual(["single2-value"], self.metadata.getraw("single2"))
            self.assertEqual(self.multi1, self.metadata.getraw("multi1"))
            self.assertEqual(self.multi2, self.metadata.getraw("multi2"))
            self.assertEqual(self.multi3, self.metadata.getraw("multi3"))
            self.assertEqual(["hidden-value"], self.metadata.getraw("~hidden"))

        def test_metadata_set_all_values_as_string(self):
            for val in (0, 2, True):
                str_val = str(val)
                self.metadata.set('val1', val)
                self.assertEqual([str_val], self.metadata.getraw("val1"))
                self.metadata['val2'] = val
                self.assertEqual([str_val], self.metadata.getraw("val2"))
                del self.metadata['val3']
                self.metadata.add('val3', val)
                self.assertEqual([str_val], self.metadata.getraw("val3"))
                del self.metadata['val4']
                self.metadata.add_unique('val4', val)
                self.assertEqual([str_val], self.metadata.getraw("val4"))

        def test_metadata_get(self):
            self.assertEqual("single1-value", self.metadata["single1"])
            self.assertEqual("single1-value", self.metadata.get("single1"))
            self.assertEqual(["single1-value"], self.metadata.getall("single1"))
            self.assertEqual(["single1-value"], self.metadata.getraw("single1"))

            self.assertEqual(MULTI_VALUED_JOINER.join(self.multi1), self.metadata["multi1"])
            self.assertEqual(MULTI_VALUED_JOINER.join(self.multi1), self.metadata.get("multi1"))
            self.assertEqual(self.multi1, self.metadata.getall("multi1"))
            self.assertEqual(self.multi1, self.metadata.getraw("multi1"))

            self.assertEqual("", self.metadata["nonexistent"])
            self.assertEqual(None, self.metadata.get("nonexistent"))
            self.assertEqual([], self.metadata.getall("nonexistent"))
            self.assertRaises(KeyError, self.metadata.getraw, "nonexistent")

            self.assertEqual(self.metadata._store.items(), self.metadata.rawitems())
            metadata_items = [(x, z) for (x, y) in self.metadata.rawitems() for z in y]
            self.assertEqual(metadata_items, list(self.metadata.items()))

        def test_metadata_unset(self):
            self.metadata.unset("single1")
            self.assertNotIn("single1", self.metadata)
            self.assertNotIn("single1", self.metadata.deleted_tags)
            self.metadata.unset('unknown_tag')

        def test_metadata_pop(self):
            self.metadata.pop("single1")
            self.assertNotIn("single1", self.metadata)
            self.assertIn("single1", self.metadata.deleted_tags)
            self.metadata.pop('unknown_tag')

        def test_metadata_delete(self):
            del self.metadata["single1"]
            self.assertNotIn("single1", self.metadata)
            self.assertIn("single1", self.metadata.deleted_tags)

        def test_metadata_legacy_delete(self):
            self.metadata.delete("single1")
            self.assertNotIn("single1", self.metadata)
            self.assertIn("single1", self.metadata.deleted_tags)

        def test_metadata_implicit_delete(self):
            self.metadata["single2"] = ""
            self.assertNotIn("single2", self.metadata)
            self.assertIn("single2", self.metadata.deleted_tags)

            self.metadata["unknown"] = ""
            self.assertNotIn("unknown", self.metadata)
            self.assertNotIn("unknown", self.metadata.deleted_tags)

        def test_metadata_undelete(self):
            self.metadata.delete("single1")
            self.assertNotIn("single1", self.metadata)
            self.assertIn("single1", self.metadata.deleted_tags)

            self.metadata["single1"] = "value1"
            self.assertIn("single1", self.metadata)
            self.assertNotIn("single1", self.metadata.deleted_tags)

        def test_normalize_tag(self):
            self.assertEqual('sometag', Metadata.normalize_tag('sometag'))
            self.assertEqual('sometag', Metadata.normalize_tag('sometag:'))
            self.assertEqual('sometag', Metadata.normalize_tag('sometag::'))
            self.assertEqual('sometag:desc', Metadata.normalize_tag('sometag:desc'))

        def test_metadata_tag_trailing_colon(self):
            self.metadata['tag:'] = 'Foo'
            self.assertIn('tag', self.metadata)
            self.assertIn('tag:', self.metadata)
            self.assertEqual('Foo', self.metadata['tag'])
            self.assertEqual('Foo', self.metadata['tag:'])
            del self.metadata['tag']
            self.assertNotIn('tag', self.metadata)
            self.assertNotIn('tag:', self.metadata)

        def test_metadata_copy(self):
            m = Metadata()
            m["old"] = "old-value"
            self.metadata.delete("single1")
            m.copy(self.metadata)
            self.assertEqual(self.metadata._store, m._store)
            self.assertEqual(self.metadata.deleted_tags, m.deleted_tags)
            self.assertEqual(self.metadata.length, m.length)
            self.assertEqual(self.metadata.images, m.images)

        def test_metadata_copy_without_images(self):
            m = Metadata()
            m.copy(self.metadata, copy_images=False)
            self.assertEqual(self.metadata._store, m._store)
            self.assertEqual(self.metadata.deleted_tags, m.deleted_tags)
            self.assertEqual(self.metadata.length, m.length)
            self.assertEqual(ImageList(), m.images)

        def test_metadata_init_with_existing_metadata(self):
            self.metadata.delete("single1")
            cover = CoverArtImage(url='file://file1', data=create_fake_png(b'a'))
            self.metadata.images.append(cover)
            m = Metadata(self.metadata)
            self.assertEqual(self.metadata.length, m.length)
            self.assertEqual(self.metadata.deleted_tags, m.deleted_tags)
            self.assertEqual(self.metadata.images, m.images)
            self.assertEqual(self.metadata._store, m._store)

        def test_metadata_update(self):
            m = Metadata()
            m["old"] = "old-value"
            self.metadata.delete("single1")
            cover = CoverArtImage(url='file://file1', data=create_fake_png(b'a'))
            self.metadata.images.append(cover)
            m.update(self.metadata)
            self.assertIn("old", m)
            self.assertNotIn("single1", m)
            self.assertIn("single1", m.deleted_tags)
            self.assertEqual("single2-value", m["single2"])
            self.assertEqual(self.metadata.length, m.length)
            self.assertEqual(self.metadata.deleted_tags, m.deleted_tags)
            self.assertEqual(self.metadata.images, m.images)
            self.metadata["old"] = "old-value"
            self.assertEqual(self.metadata._store, m._store)

        def test_metadata_diff(self):
            m1 = Metadata({
                "foo1": "bar1",
                "foo2": "bar2",
                "foo3": "bar3",
            })
            m2 = Metadata(m1)
            m1["foo1"] = "baz"
            del m1["foo2"]
            diff = m1.diff(m2)
            self.assertEqual({"foo1": "baz"}, diff)
            self.assertEqual(set(["foo2"]), diff.deleted_tags)

        def test_metadata_clear(self):
            self.metadata.clear()
            self.assertEqual(0, len(self.metadata))

        def test_metadata_clear_deleted(self):
            self.metadata.delete("single1")
            self.assertIn("single1", self.metadata.deleted_tags)
            self.metadata.clear_deleted()
            self.assertNotIn("single1", self.metadata.deleted_tags)

        def test_metadata_applyfunc(self):
            def func(x):
                return x[1:]
            self.metadata.apply_func(func)

            self.assertEqual("ingle1-value", self.metadata["single1"])
            self.assertEqual("ingle1-value", self.metadata.get("single1"))
            self.assertEqual(["ingle1-value"], self.metadata.getall("single1"))

            self.assertEqual(MULTI_VALUED_JOINER.join(map(func, self.multi1)), self.metadata["multi1"])
            self.assertEqual(MULTI_VALUED_JOINER.join(map(func, self.multi1)), self.metadata.get("multi1"))
            self.assertEqual(list(map(func, self.multi1)), self.metadata.getall("multi1"))

        def test_metadata_applyfunc_preserve_tags(self):
            self.assertTrue(len(PRESERVED_TAGS) > 0)
            m = Metadata()
            m[PRESERVED_TAGS[0]] = 'value1'
            m['not_preserved'] = 'value2'

            def func(x):
                return x[1:]
            m.apply_func(func)

            self.assertEqual("value1", m[PRESERVED_TAGS[0]])
            self.assertEqual("alue2", m['not_preserved'])

        def test_metadata_applyfunc_delete_tags(self):
            def func(x):
                return None
            metadata = Metadata(self.metadata)
            metadata.apply_func(func)
            self.assertEqual(0, len(metadata.rawitems()))
            self.assertEqual(self.metadata.keys(), metadata.deleted_tags)

        def test_length_score(self):
            results = [(20000, 0, 0.333333333333),
                       (20000, 10000, 0.666666666667),
                       (20000, 20000, 1.0),
                       (20000, 30000, 0.666666666667),
                       (20000, 40000, 0.333333333333),
                       (20000, 50000, 0.0),
                       (20000, None, 0.0),
                       (None, 2000, 0.0),
                       (None, None, 0.0)]
            for (a, b, expected) in results:
                actual = Metadata.length_score(a, b)
                self.assertAlmostEqual(expected, actual,
                                       msg="a={a}, b={b}".format(a=a, b=b))

        def test_compare_is_equal(self):
            m1 = Metadata()
            m1["title"] = "title1"
            m1["tracknumber"] = "2"
            m1.length = 360
            m2 = Metadata()
            m2["title"] = "title1"
            m2["tracknumber"] = "2"
            m2.length = 360
            self.assertEqual(m1.compare(m2), m2.compare(m1))
            self.assertEqual(m1.compare(m2), 1)

        def test_compare_with_ignored(self):
            m1 = Metadata()
            m1["title"] = "title1"
            m1["tracknumber"] = "2"
            m1.length = 360
            m2 = Metadata()
            m2["title"] = "title1"
            m2["tracknumber"] = "3"
            m2.length = 300
            self.assertNotEqual(m1.compare(m2), 1)
            self.assertEqual(m1.compare(m2, ignored=['tracknumber', '~length']), 1)

        def test_compare_lengths(self):
            m1 = Metadata()
            m1.length = 360
            m2 = Metadata()
            m2.length = 300
            self.assertAlmostEqual(m1.compare(m2), 0.998)

        def test_compare_tracknumber_difference(self):
            m1 = Metadata()
            m1["tracknumber"] = "1"
            m2 = Metadata()
            m2["tracknumber"] = "2"
            m3 = Metadata()
            m3["tracknumber"] = "2"
            self.assertEqual(m1.compare(m2), 0)
            self.assertEqual(m2.compare(m3), 1)

        def test_compare_discnumber_difference(self):
            m1 = Metadata()
            m1["discnumber"] = "1"
            m2 = Metadata()
            m2["discnumber"] = "2"
            m3 = Metadata()
            m3["discnumber"] = "2"
            self.assertEqual(m1.compare(m2), 0)
            self.assertEqual(m2.compare(m3), 1)

        def test_compare_deleted(self):
            m1 = Metadata()
            m1["artist"] = "TheArtist"
            m1["title"] = "title1"
            m2 = Metadata()
            m2["artist"] = "TheArtist"
            m2.delete("title")
            self.assertTrue(m1.compare(m2) < 1)

        def test_strip_whitespace(self):
            m1 = Metadata()
            m1["artist"] = "  TheArtist  "
            m1["title"] = "\t\u00A0  tit le1 \r\n"
            m1["genre"] = " \t"
            m1.strip_whitespace()
            self.assertEqual(m1["artist"], "TheArtist")
            self.assertEqual(m1["title"], "tit le1")

        def test_metadata_mapping_init(self):
            d = {'a': 'b', 'c': 2, 'd': ['x', 'y'], 'x': '', 'z': {'u', 'w'}}
            deleted_tags = set('c')
            m = Metadata(d, deleted_tags=deleted_tags, length=1234)
            self.assertIn('a', m)
            self.assertEqual(m.getraw('a'), ['b'])
            self.assertEqual(m['d'], MULTI_VALUED_JOINER.join(d['d']))
            self.assertNotIn('c', m)
            self.assertNotIn('length', m)
            self.assertIn('c', m.deleted_tags)
            self.assertEqual(m.length, 1234)

        def test_metadata_mapping_init_zero(self):
            m = Metadata(tag1='a', tag2=0, tag3='', tag4=None)
            m['tag5'] = 0
            m['tag1'] = ''
            self.assertIn('tag1', m.deleted_tags)
            self.assertEqual(m['tag2'], '0')
            self.assertNotIn('tag3', m)
            self.assertNotIn('tag4', m)
            self.assertEqual(m['tag5'], '0')

        def test_metadata_mapping_del(self):
            m = self.metadata_d1
            self.assertEqual(m.getraw('a'), ['b'])
            self.assertNotIn('a', m.deleted_tags)

            self.assertNotIn('x', m.deleted_tags)
            self.assertRaises(KeyError, m.getraw, 'x')

            del m['a']
            self.assertRaises(KeyError, m.getraw, 'a')
            self.assertIn('a', m.deleted_tags)

            # NOTE: historic behavior of Metadata.delete()
            # an attempt to delete an non-existing tag, will add it to the list
            # of deleted tags
            # so this will not raise a KeyError
            # as is it differs from dict or even defaultdict behavior
            del m['unknown']
            self.assertIn('unknown', m.deleted_tags)

        def test_metadata_mapping_iter(self):
            self.assertEqual(set(self.metadata_d1), {'a', 'c', 'd'})

        def test_metadata_mapping_keys(self):
            self.assertEqual(set(self.metadata_d1.keys()), {'a', 'c', 'd'})

        def test_metadata_mapping_values(self):
            self.assertEqual(set(self.metadata_d1.values()), {'b', '2', 'x; y'})

        def test_metadata_mapping_len(self):
            m = self.metadata_d1
            self.assertEqual(len(m), 3)
            del m['x']
            self.assertEqual(len(m), 3)
            del m['c']
            self.assertEqual(len(m), 2)

        def _check_mapping_update(self, m):
            self.assertEqual(m['a'], 'b')
            self.assertEqual(m['c'], '3')
            self.assertEqual(m.getraw('d'), ['u', 'w'])
            self.assertEqual(m['x'], '')
            self.assertIn('x', m.deleted_tags)

        def test_metadata_mapping_update(self):
            # update from Metadata
            m = self.metadata_d2
            m2 = self.metadata_d3

            del m2['x']
            m.update(m2)
            self._check_mapping_update(m)

        def test_metadata_mapping_update_dict(self):
            # update from dict
            m = self.metadata_d2

            d2 = {'c': 3, 'd': ['u', 'w'], 'x': ''}

            m.update(d2)
            self._check_mapping_update(m)

        def test_metadata_mapping_update_tuple(self):
            # update from tuple
            m = self.metadata_d2

            d2 = (('c', 3), ('d', ['u', 'w']), ('x', ''))

            m.update(d2)
            self._check_mapping_update(m)

        def test_metadata_mapping_update_dictlike(self):
            # update from kwargs
            m = self.metadata_d2

            m.update(c=3, d=['u', 'w'], x='')
            self._check_mapping_update(m)

        def test_metadata_mapping_update_noparam(self):
            # update without parameter
            m = self.metadata_d2

            self.assertRaises(TypeError, m.update)
            self.assertEqual(m['a'], 'b')

        def test_metadata_mapping_update_intparam(self):
            # update without parameter
            m = self.metadata_d2

            self.assertRaises(TypeError, m.update, 123)

        def test_metadata_mapping_update_strparam(self):
            # update without parameter
            m = self.metadata_d2

            self.assertRaises(ValueError, m.update, 'abc')

        def test_metadata_mapping_update_kw(self):
            m = Metadata(tag1='a', tag2='b')
            m.update(tag1='c')
            self.assertEqual(m['tag1'], 'c')
            self.assertEqual(m['tag2'], 'b')
            m.update(tag2='')
            self.assertIn('tag2', m.deleted_tags)

        def test_metadata_mapping_update_kw_del(self):
            m = Metadata(tag1='a', tag2='b')
            del m['tag1']

            m2 = Metadata(tag1='c', tag2='d')
            del m2['tag2']

            m.update(m2)
            self.assertEqual(m['tag1'], 'c')
            self.assertNotIn('tag2', m)
            self.assertNotIn('tag1', m.deleted_tags)
            self.assertIn('tag2', m.deleted_tags)

        def test_metadata_mapping_images(self):
            image1 = create_image(b'A', comment='A')
            image2 = create_image(b'B', comment='B')

            m1 = Metadata(a='b', length=1234, images=[image1])
            self.assertEqual(m1.images[0], image1)
            self.assertEqual(len(m1), 2)  # one tag, one image

            m1.images.append(image2)
            self.assertEqual(m1.images[1], image2)

            m1.images.pop(0)
            self.assertEqual(m1.images[0], image2)

            m2 = Metadata(a='c', length=4567, images=[image1])
            m1.update(m2)
            self.assertEqual(m1.images[0], image1)

            m1.images.pop(0)
            self.assertEqual(len(m1), 1)  # one tag, zero image
            self.assertFalse(m1.images)

        def test_metadata_mapping_iterable(self):
            m = Metadata(tag_tuple=('a', 0))
            m['tag_set'] = {'c', 'd'}
            m['tag_dict'] = {'e': 1, 'f': 2}
            m['tag_str'] = 'gh'
            self.assertIn('0', m.getraw('tag_tuple'))
            self.assertIn('c', m.getraw('tag_set'))
            self.assertIn('e', m.getraw('tag_dict'))
            self.assertIn('gh', m.getraw('tag_str'))

        def test_compare_to_release(self):
            release = load_test_json('release.json')
            metadata = Metadata()
            release_to_metadata(release, metadata)
            match = metadata.compare_to_release(release, Cluster.comparison_weights)
            self.assertEqual(1.0, match.similarity)
            self.assertEqual(release, match.release)

        def test_compare_to_release_with_score(self):
            release = load_test_json('release.json')
            metadata = Metadata()
            release_to_metadata(release, metadata)
            for score, sim in ((42, 0.42), ('42', 0.42), ('foo', 1.0), (None, 1.0)):
                release['score'] = score
                match = metadata.compare_to_release(release, Cluster.comparison_weights)
                self.assertEqual(sim, match.similarity)

        def test_compare_to_release_parts_totaltracks(self):
            release = load_test_json('release_multidisc.json')
            metadata = Metadata()
            weights = {"totaltracks": 30}
            release_to_metadata(release, metadata)
            for totaltracks, sim in ((4, 1.0), (3, 1.0), (2, 0.3), (5, 0.0)):
                metadata['totaltracks'] = totaltracks
                parts = metadata.compare_to_release_parts(release, weights)
                self.assertIn((sim, 30), parts)

        def test_compare_to_release_parts_totalalbumtracks(self):
            release = load_test_json('release_multidisc.json')
            metadata = Metadata()
            weights = {"totalalbumtracks": 30}
            release_to_metadata(release, metadata)
            for totaltracks, sim in ((7, 1.0), (6, 0.3), (8, 0.0)):
                metadata['~totalalbumtracks'] = totaltracks
                parts = metadata.compare_to_release_parts(release, weights)
                self.assertIn((sim, 30), parts)

        def test_compare_to_release_parts_totalalbumtracks_totaltracks_fallback(self):
            release = load_test_json('release_multidisc.json')
            metadata = Metadata()
            weights = {"totalalbumtracks": 30}
            release_to_metadata(release, metadata)
            for totaltracks, sim in ((7, 1.0), (6, 0.3), (8, 0.0)):
                metadata['totaltracks'] = totaltracks
                parts = metadata.compare_to_release_parts(release, weights)
                self.assertIn((sim, 30), parts)

        def test_trackcount_score(self):
            self.assertEqual(1.0, trackcount_score(5, 5))
            self.assertEqual(0.0, trackcount_score(6, 5))
            self.assertEqual(0.3, trackcount_score(4, 5))

        def test_weights_from_release_type_scores(self):
            release = load_test_json('release.json')
            parts = []
            weights_from_release_type_scores(parts, release, {'Album': 0.75}, 666)
            self.assertEqual(
                parts[0],
                (0.75, 666)
            )
            weights_from_release_type_scores(parts, release, {}, 666)
            self.assertEqual(
                parts[1],
                (0.5, 666)
            )

        def test_weights_from_release_type_scores_no_type(self):
            release = load_test_json('release_no_type.json')
            parts = []
            weights_from_release_type_scores(parts, release, {'Other': 0.75}, 123)
            self.assertEqual(
                parts[0],
                (0.75, 123)
            )
            weights_from_release_type_scores(parts, release, {}, 123)
            self.assertEqual(
                parts[1],
                (0.5, 123)
            )

        def test_preferred_countries(self):
            release = load_test_json('release.json')
            parts = []
            weights_from_preferred_countries(parts, release, [], 666)
            self.assertFalse(parts)
            weights_from_preferred_countries(parts, release, ['FR'], 666)
            self.assertEqual(parts[0], (0.0, 666))
            weights_from_preferred_countries(parts, release, ['GB'], 666)
            self.assertEqual(parts[1], (1.0, 666))

        def test_preferred_formats(self):
            release = load_test_json('release.json')
            parts = []
            weights_from_preferred_formats(parts, release, [], 777)
            self.assertFalse(parts)
            weights_from_preferred_formats(parts, release, ['Digital Media'], 777)
            self.assertEqual(parts[0], (0.0, 777))
            weights_from_preferred_formats(parts, release, ['12" Vinyl'], 777)
            self.assertEqual(parts[1], (1.0, 777))

        def test_compare_to_track(self):
            track_json = load_test_json('track.json')
            track = Track(track_json['id'])
            track_to_metadata(track_json, track)
            match = track.metadata.compare_to_track(track_json, File.comparison_weights)
            self.assertEqual(1.0, match.similarity)
            self.assertEqual(track_json, match.track)

        def test_compare_to_track_with_score(self):
            track_json = load_test_json('track.json')
            track = Track(track_json['id'])
            track_to_metadata(track_json, track)
            for score, sim in ((42, 0.42), ('42', 0.42), ('foo', 1.0), (None, 1.0)):
                track_json['score'] = score
                match = track.metadata.compare_to_track(track_json, File.comparison_weights)
                self.assertEqual(sim, match.similarity)

        def test_compare_to_track_is_video(self):
            recording = load_test_json('recording_video_null.json')
            m = Metadata()
            match = m.compare_to_track(recording, {'isvideo': 1})
            self.assertEqual(1.0, match.similarity)
            m['~video'] = '1'
            match = m.compare_to_track(recording, {'isvideo': 1})
            self.assertEqual(0.0, match.similarity)
            recording['video'] = True
            match = m.compare_to_track(recording, {'isvideo': 1})
            self.assertEqual(1.0, match.similarity)

        def test_compare_to_track_full(self):
            recording = load_test_json('recording_video_null.json')
            m = Metadata({
                'artist': 'Tim Green',
                'release': 'Eastbound Silhouette',
                'date': '2022',
                'title': 'Lune',
                'totaltracks': '6',
                'albumartist': 'Tim Green',
                'tracknumber': '4',
            })
            match = m.compare_to_track(recording, File.comparison_weights)
            self.assertGreaterEqual(match.similarity, 0.8)
            self.assertEqual(recording, match.track)
            self.assertEqual(recording['releases'][0], match.release)

        def test_compare_to_track_without_releases(self):
            self.set_config_values({
                'release_type_scores': [('Compilation', 0.6), ('Other', 0.6)]
            })
            track_json = acoustid_parse_recording(load_test_json('acoustid.json'))
            track = Track(track_json['id'])
            track.metadata.update({
                'album': 'x',
                'artist': 'Ed Sheeran',
                'title': 'Nina',
            })
            track.metadata.length = 225000
            m1 = track.metadata.compare_to_track(track_json, File.comparison_weights)
            del track_json['releases']
            m2 = track.metadata.compare_to_track(track_json, File.comparison_weights)
            self.assertGreater(m1.similarity, m2.similarity,
                               'Matching score for release with recordings must be higher then for release without')


class MetadataTest(CommonTests.CommonMetadataTestCase):
    @staticmethod
    def get_metadata_object():
        return Metadata()


class MultiMetadataProxyAsMetadataTest(CommonTests.CommonMetadataTestCase):
    @staticmethod
    def get_metadata_object():
        return MultiMetadataProxy(Metadata())


class MultiMetadataProxyTest(PicardTestCase):

    def setUp(self):
        super().setUp()
        self.m1 = Metadata({
            "key1": "m1.val1",
            "key2": "m1.val2",
        })
        self.m2 = Metadata({
            "key2": "m2.val2",
            "key3": "m2.val3",
        })
        self.m3 = Metadata({
            "key2": "m3.val2",
            "key4": "m3.val4",
        })

    def test_get_attribute(self):
        mp = MultiMetadataProxy(self.m1, self.m2, self.m3)
        self.assertEqual(mp.deleted_tags, self.m1.deleted_tags)

    def test_gettitem(self):
        mp = MultiMetadataProxy(self.m1, self.m2, self.m3)
        self.assertEqual("m1.val1", mp["key1"])
        self.assertEqual("m1.val2", mp["key2"])
        self.assertEqual("m2.val3", mp["key3"])
        self.assertEqual("m3.val4", mp["key4"])

    def test_settitem(self):
        orig_m2 = Metadata(self.m2)
        mp = MultiMetadataProxy(self.m1, self.m2, self.m3)
        mp["key1"] = "foo1"
        mp["key2"] = "foo2"
        mp["key3"] = "foo3"
        mp["key4"] = "foo4"
        mp["key5"] = "foo5"
        self.assertEqual("foo1", self.m1["key1"])
        self.assertEqual("foo2", self.m1["key2"])
        self.assertEqual("foo3", self.m1["key3"])
        self.assertEqual("foo4", self.m1["key4"])
        self.assertEqual("foo5", self.m1["key5"])
        self.assertEqual(orig_m2, self.m2)

    def test_delitem(self):
        orig_m2 = Metadata(self.m2)
        mp = MultiMetadataProxy(self.m1, self.m2, self.m3)
        del mp["key2"]
        del mp["key3"]
        self.assertIn("key2", self.m1.deleted_tags)
        self.assertIn("key3", self.m1.deleted_tags)
        self.assertEqual(orig_m2, self.m2)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2006-2009, 2012 Lukáš Lalinský
# Copyright (C) 2007 Javier Kohen
# Copyright (C) 2008-2011, 2014-2015, 2018-2023 Philipp Wolfer
# Copyright (C) 2009 Carlin Mangar
# Copyright (C) 2009 Nikolai Prokoschenko
# Copyright (C) 2011-2012 Michael Wiencek
# Copyright (C) 2012 Chad Wilson
# Copyright (C) 2012 stephen
# Copyright (C) 2012, 2014, 2017, 2021 Wieland Hoffmann
# Copyright (C) 2013-2014, 2017-2022 Laurent Monin
# Copyright (C) 2014, 2017, 2021 Sophist-UK
# Copyright (C) 2016-2017 Sambhav Kothari
# Copyright (C) 2016-2017 Ville Skyttä
# Copyright (C) 2017-2018 Antonio Larrosa
# Copyright (C) 2018 Calvin Walton
# Copyright (C) 2018 virusMac
# Copyright (C) 2020-2022 Bob Swift
# Copyright (C) 2021 Adam James
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from collections import namedtuple
import datetime
from functools import reduce
from inspect import getfullargspec
import operator
import re

from picard.const.countries import RELEASE_COUNTRIES
from picard.metadata import MULTI_VALUED_JOINER
from picard.script.parser import (
    MultiValue,
    ScriptParser,
    ScriptRuntimeError,
    normalize_tagname,
)
from picard.util import (
    pattern_as_regex,
    titlecase,
    uniqify,
)


try:
    from markdown import markdown
except ImportError:
    markdown = None


Bound = namedtuple('Bound', ['lower', 'upper'])


class FunctionRegistryItem:
    def __init__(self, function, eval_args, argcount, documentation=None,
                 name=None, module=None):
        self.function = function
        self.eval_args = eval_args
        self.argcount = argcount
        self.documentation = documentation
        self.name = name
        self.module = module

    def __repr__(self):
        return '{classname}({me.function}, {me.eval_args}, {me.argcount}, {doc})'.format(
            classname=self.__class__.__name__,
            me=self,
            doc='"""{0}"""'.format(self.documentation) if self.documentation else None
        )

    def _postprocess(self, data, postprocessor):
        if postprocessor is not None:
            data = postprocessor(data, function=self)
        return data

    def markdowndoc(self, postprocessor=None):
        if self.documentation is not None:
            ret = _(self.documentation)
        else:
            ret = ''
        return self._postprocess(ret, postprocessor)

    def htmldoc(self, postprocessor=None):
        if markdown is not None:
            ret = markdown(self.markdowndoc())
        else:
            ret = ''
        return self._postprocess(ret, postprocessor)


def register_script_function(function, name=None, eval_args=True,
                             check_argcount=True, documentation=None):
    """Registers a script function. If ``name`` is ``None``,
    ``function.__name__`` will be used.
    If ``eval_args`` is ``False``, the arguments will not be evaluated before being
    passed to ``function``.
    If ``check_argcount`` is ``False`` the number of arguments passed to the
    function will not be verified."""

    args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations = getfullargspec(function)

    required_kwonlyargs = len(kwonlyargs)
    if kwonlydefaults is not None:
        required_kwonlyargs -= len(kwonlydefaults.keys())
    if required_kwonlyargs:
        raise TypeError("Functions with required keyword-only parameters are not supported")

    args = len(args) - 1  # -1 for the parser
    varargs = varargs is not None
    defaults = len(defaults) if defaults else 0

    argcount = Bound(args - defaults, args if not varargs else None)

    if name is None:
        name = function.__name__
    ScriptParser._function_registry.register(
        function.__module__,
        (
            name,
            FunctionRegistryItem(
                function,
                eval_args,
                argcount if argcount and check_argcount else False,
                documentation=documentation,
                name=name,
                module=function.__module__,
            )
        )
    )


def script_function(name=None, eval_args=True, check_argcount=True, prefix='func_', documentation=None):
    """Decorator helper to register script functions

    It calls ``register_script_function()`` and share same arguments
    Extra optional arguments:
        ``prefix``: define the prefix to be removed from defined function to name script function
                    By default, ``func_foo`` will create ``foo`` script function

    Example:
        @script_function(eval_args=False)
        def func_myscriptfunc():
            ...
    """
    def script_function_decorator(func):
        fname = func.__name__
        if name is None and prefix and fname.startswith(prefix):
            sname = fname[len(prefix):]
        else:
            sname = name
        register_script_function(
            func,
            name=sname,
            eval_args=eval_args,
            check_argcount=check_argcount,
            documentation=documentation
        )
        return func
    return script_function_decorator


def _compute_int(operation, *args):
    return str(reduce(operation, map(int, args)))


def _compute_logic(operation, *args):
    return operation(args)


@script_function(eval_args=False, documentation=N_(
    """`$if(if,then,else)`

If `if` is not empty, it returns `then`, otherwise it returns `else`."""
))
def func_if(parser, _if, _then, _else=None):
    if _if.eval(parser):
        return _then.eval(parser)
    elif _else:
        return _else.eval(parser)
    return ''


@script_function(eval_args=False, documentation=N_(
    """`$if2(a1,a2,a3,…)`

Returns first non empty argument."""
))
def func_if2(parser, *args):
    for arg in args:
        arg = arg.eval(parser)
        if arg:
            return arg
    return ''


@script_function(eval_args=False, documentation=N_(
    """`$noop(…)`

Does nothing (useful for comments or disabling a block of code)."""
))
def func_noop(parser, *args):
    return ''


@script_function(documentation=N_(
    """`$left(text,number)`

Returns the first `number` characters from `text`."""
))
def func_left(parser, text, length):
    try:
        return text[:int(length)]
    except ValueError:
        return ''


@script_function(documentation=N_(
    """`$right(text,number)`

Returns the last `number` characters from `text`."""
))
def func_right(parser, text, length):
    try:
        return text[-int(length):]
    except ValueError:
        return ''


@script_function(documentation=N_(
    """`$lower(text)`

Returns `text` in lower case."""
))
def func_lower(parser, text):
    return text.lower()


@script_function(documentation=N_(
    """`$upper(text)`

Returns `text` in upper case."""
))
def func_upper(parser, text):
    return text.upper()


@script_function(documentation=N_(
    """`$pad(text,length,char)`

Pads the `text` to the `length` provided by adding as many copies of `char` as
    needed to the **beginning** of the string."""
))
def func_pad(parser, text, length, char):
    try:
        return char * (int(length) - len(text)) + text
    except ValueError:
        return ''


@script_function(documentation=N_(
    """`$strip(text)`

Replaces all whitespace in `text` with a single space, and removes leading and trailing spaces.
Whitespace characters include multiple consecutive spaces, and various other unicode characters."""
))
def func_strip(parser, text):
    return re.sub(r"\s+", " ", text).strip()


@script_function(documentation=N_(
    """`$replace(text,search,replace)`

Replaces occurrences of `search` in `text` with value of `replace` and returns the resulting string."""
))
def func_replace(parser, text, old, new):
    return text.replace(old, new)


@script_function(eval_args=False, documentation=N_(
    """`$replacemulti(name,search,replace,separator="; ")`

Replaces occurrences of `search` with `replace` in the multi-value variable `name`.
Empty elements are automatically removed.

Example:

    $replacemulti(%genre%,Idm,IDM)
"""
))
def func_replacemulti(parser, multi, search, replace, separator=MULTI_VALUED_JOINER):
    if not multi or not search or replace is None or not separator:
        return multi.eval(parser)

    search = search.eval(parser)
    replace = replace.eval(parser)
    multi_value = MultiValue(parser, multi, separator)
    for n, value in enumerate(multi_value):
        if value == search:
            multi_value[n] = replace

    return str(multi_value)


@script_function(documentation=N_(
    """`$in(x,y)`

Returns true, if `x` contains `y`."""
))
def func_in(parser, text, needle):
    if needle in text:
        return '1'
    else:
        return ''


@script_function(eval_args=False, documentation=N_(
    """`$inmulti(%x%,y)`

Returns true if multi-value variable `x` contains exactly `y` as one of its values.

_Since Picard 1.0_"""
))
def func_inmulti(parser, haystack, needle, separator=MULTI_VALUED_JOINER):
    """Searches for ``needle`` in ``haystack``, supporting a list variable for
       ``haystack``. If a string is used instead, then a ``separator`` can be
       used to split it. In both cases, it returns true if the resulting list
       contains exactly ``needle`` as a member."""

    needle = needle.eval(parser)
    return func_in(parser, MultiValue(parser, haystack, separator), needle)


@script_function(documentation=N_(
    """`$rreplace(text,pattern,replace)`

[Regular expression](https://docs.python.org/3/library/re.html#regular-expression-syntax) replace."""
))
def func_rreplace(parser, text, old, new):
    try:
        return re.sub(old, new, text)
    except re.error:
        return text


@script_function(documentation=N_(
    """`$rsearch(text,pattern)`

[Regular expression](https://docs.python.org/3/library/re.html#regular-expression-syntax) search.
    This function will return the first matching group."""
))
def func_rsearch(parser, text, pattern):
    try:
        match = re.search(pattern, text)
    except re.error:
        return ''
    if match:
        try:
            return match.group(1)
        except IndexError:
            return match.group(0)
    return ''


@script_function(documentation=N_(
    """`$num(number,length)`

Returns `number` formatted to `length` digits (maximum 20)."""
))
def func_num(parser, text, length):
    try:
        format_ = "%%0%dd" % max(0, min(int(length), 20))
    except ValueError:
        return ''
    try:
        value = int(text)
    except ValueError:
        value = 0
    return format_ % value


@script_function(documentation=N_(
    """`$unset(name)`

Unsets the variable `name`.
Allows for wildcards to unset certain tags (works with "performer:\\*", "comment:\\*", and "lyrics:\\*").
For example `$unset(performer:*)` would unset all performer tags."""
))
def func_unset(parser, name):
    name = normalize_tagname(name)
    # Allow wild-card unset for certain keys
    if name in {'performer:*', 'comment:*', 'lyrics:*'}:
        name = name[:-1]
        for key in list(parser.context.keys()):
            if key.startswith(name):
                parser.context.unset(key)
        return ''
    parser.context.unset(name)
    return ''


@script_function(documentation=N_(
    """`$delete(name)`

Unsets the variable `name` and marks the tag for deletion.
This is similar to `$unset(name)` but also marks the tag for deletion. E.g.
    running `$delete(genre)` will actually remove the genre tag from a file when
    saving.

_Since Picard 2.1_"""
))
def func_delete(parser, name):
    parser.context.delete(normalize_tagname(name))
    return ''


@script_function(documentation=N_(
    """`$set(name,value)`

Sets the variable `name` to `value`.

Note: To create a variable which can be used for the file naming string, but
    which will not be written as a tag in the file, prefix the variable name
    with an underscore. `%something%` will create a "something" tag;
    `%_something%` will not."""
))
def func_set(parser, name, value):
    if value:
        parser.context[normalize_tagname(name)] = value
    else:
        func_unset(parser, name)
    return ''


@script_function(documentation=N_(
    """`$setmulti(name,value,separator="; ")`

Sets the variable `name` to `value`, using the separator (or "; " if not passed)
    to coerce the value back into a proper multi-valued tag. This can be used to
    operate on multi-valued tags as a string, and then set them back as proper
    multi-valued tags.

Example:

    $setmulti(genre,$lower(%genre%))

_Since Picard 1.0_"""
))
def func_setmulti(parser, name, value, separator=MULTI_VALUED_JOINER):
    return func_set(parser, name, value.split(separator) if value and separator else value)


@script_function(documentation=N_(
    """`$get(name)`

Returns the variable `name` (equivalent to `%name%`)."""
))
def func_get(parser, name):
    """Returns the variable ``name`` (equivalent to ``%name%``)."""
    return parser.context.get(normalize_tagname(name), "")


@script_function(documentation=N_(
    """`$copy(new,old)`

Copies metadata from variable `old` to `new`.
The difference between `$set(new,%old%)` is that `$copy(new,old)` copies
    multi-value variables without flattening them.

_Since Picard 0.9_"""
))
def func_copy(parser, new, old):
    new = normalize_tagname(new)
    old = normalize_tagname(old)
    parser.context[new] = parser.context.getall(old)[:]
    return ''


@script_function(documentation=N_(
    """`$copymerge(new,old[,keep_duplicates])`

Merges metadata from variable `old` into `new`, removing duplicates and
    appending to the end, so retaining the original ordering. Like `$copy`, this
    will also copy multi-valued variables without flattening them.

If `keep_duplicates` is set, then the duplicates will not be removed from the result.

_Since Picard 1.0_"""
))
def func_copymerge(parser, new, old, keep_duplicates=False):
    new = normalize_tagname(new)
    old = normalize_tagname(old)
    newvals = parser.context.getall(new)
    oldvals = parser.context.getall(old)
    parser.context[new] = newvals + oldvals if keep_duplicates else uniqify(newvals + oldvals)
    return ''


@script_function(documentation=N_(
    """`$trim(text[,char])`

Trims all leading and trailing whitespaces from `text`.
    The optional second parameter `char` specifies the character to trim."""
))
def func_trim(parser, text, char=None):
    if char:
        return text.strip(char)
    else:
        return text.strip()


@script_function(documentation=N_(
    """`$add(x,y,…)`

Add `y` to `x`.
Can be used with an arbitrary number of arguments.

Example:

    $add(x,y,z) = ((x + y) + z)
"""
))
def func_add(parser, x, y, *args):
    try:
        return _compute_int(operator.add, x, y, *args)
    except ValueError:
        return ''


@script_function(documentation=N_(
    """`$sub(x,y,…)`

Subtracts `y` from `x`.
Can be used with an arbitrary number of arguments.

Example:

    $sub(x,y,z) = ((x - y) - z)
"""
))
def func_sub(parser, x, y, *args):
    try:
        return _compute_int(operator.sub, x, y, *args)
    except ValueError:
        return ''


@script_function(documentation=N_(
    """`$div(x,y,…)`

Divides `x` by `y`.
Can be used with an arbitrary number of arguments.

Example:

    $div(x,y,z) = ((x / y) / z)
"""
))
def func_div(parser, x, y, *args):
    try:
        return _compute_int(operator.floordiv, x, y, *args)
    except ValueError:
        return ''
    except ZeroDivisionError:
        return ''


@script_function(documentation=N_(
    """`$mod(x,y,…)`

Returns the remainder of `x` divided by `y`.
Can be used with an arbitrary number of arguments.

Example:

    $mod(x,y,z) = ((x % y) % z)
"""
))
def func_mod(parser, x, y, *args):
    try:
        return _compute_int(operator.mod, x, y, *args)
    except (ValueError, ZeroDivisionError):
        return ''


@script_function(documentation=N_(
    """`$mul(x,y,…)`

Multiplies `x` by `y`.
Can be used with an arbitrary number of arguments.

Example:

    $mul(x,y,z) = ((x * y) * z)
"""
))
def func_mul(parser, x, y, *args):
    try:
        return _compute_int(operator.mul, x, y, *args)
    except ValueError:
        return ''


@script_function(documentation=N_(
    """`$or(x,y,…)`

Returns true if either `x` or `y` not empty.
    Can be used with an arbitrary number of arguments.
    The result is true if ANY of the arguments is not empty."""
))
def func_or(parser, x, y, *args):
    if _compute_logic(any, x, y, *args):
        return '1'
    else:
        return ''


@script_function(documentation=N_(
    """`$and(x,y,…)`

Returns true if both `x` and `y` are not empty.
    Can be used with an arbitrary number of arguments.
    The result is true if ALL of the arguments are not empty."""
))
def func_and(parser, x, y, *args):
    if _compute_logic(all, x, y, *args):
        return '1'
    else:
        return ''


@script_function(documentation=N_(
    """`$not(x)`

Returns true if `x` is empty."""
))
def func_not(parser, x):
    if not x:
        return '1'
    else:
        return ''


@script_function(documentation=N_(
    """`$eq(x,y)`

Returns true if `x` equals `y`."""
))
def func_eq(parser, x, y):
    if x == y:
        return '1'
    else:
        return ''


@script_function(documentation=N_(
    """`$ne(x,y)`

Returns true if `x` does not equal `y`."""
))
def func_ne(parser, x, y):
    if x != y:
        return '1'
    else:
        return ''


def _cmp(op, x, y, _type):
    """Compare x vs y using op method and specified _type
       op is expected to be a method from operator module
    """
    if not _type:
        _type = 'auto'
    _typer = None
    if _type == 'auto':
        _type = 'text'
        for _test_type in (int, float):
            try:
                _type = (_test_type(x), _test_type(y))
                _type = _test_type.__name__
                break
            except ValueError:
                pass
    if _type == 'text':
        return '1' if op(x, y) else ""
    elif _type == 'nocase':
        return '1' if op(x.lower(), y.lower()) else ""
    elif _type == 'float':
        _typer = float
    elif _type == 'int':
        _typer = int
    if _typer is not None:
        try:
            if op(_typer(x), _typer(y)):
                return '1'
        except ValueError:
            pass
    return ''


@script_function(documentation=N_(
    """`$lt(x,y[,type])`

Returns true if `x` is less than `y` using the comparison specified in `type`.
Possible values of `type` are "int" (integer), "float" (floating point), "text"
(case-sensitive text), "nocase" (case-insensitive text) and "auto" (automatically
determine the type of arguments provided), with "auto" used as the default
comparison method if `type` is not specified.  The "auto" type will use the
first type that applies to both arguments in the following order of preference:
"int", "float" and "text"."""
))
def func_lt(parser, x, y, _type=None):
    return _cmp(operator.lt, x, y, _type)


@script_function(documentation=N_(
    """`$lte(x,y[,type])`

Returns true if `x` is less than or equal to `y` using the comparison specified in `type`.
Possible values of `type` are "int" (integer), "float" (floating point), "text"
(case-sensitive text), "nocase" (case-insensitive text) and "auto" (automatically
determine the type of arguments provided), with "auto" used as the default
comparison method if `type` is not specified.  The "auto" type will use the
first type that applies to both arguments in the following order of preference:
"int", "float" and "text"."""
))
def func_lte(parser, x, y, _type=None):
    return _cmp(operator.le, x, y, _type)


@script_function(documentation=N_(
    """`$gt(x,y[,type])`

Returns true if `x` is greater than `y` using the comparison specified in `type`.
Possible values of `type` are "int" (integer), "float" (floating point), "text"
(case-sensitive text), "nocase" (case-insensitive text) and "auto" (automatically
determine the type of arguments provided), with "auto" used as the default
comparison method if `type` is not specified.  The "auto" type will use the
first type that applies to both arguments in the following order of preference:
"int", "float" and "text"."""
))
def func_gt(parser, x, y, _type=None):
    return _cmp(operator.gt, x, y, _type)


@script_function(documentation=N_(
    """`$gte(x,y[,type])`

Returns true if `x` is greater than or equal to `y` using the comparison specified in `type`.
Possible values of `type` are "int" (integer), "float" (floating point), "text"
(case-sensitive text), "nocase" (case-insensitive text) and "auto" (automatically
determine the type of arguments provided), with "auto" used as the default
comparison method if `type` is not specified.  The "auto" type will use the
first type that applies to both arguments in the following order of preference:
"int", "float" and "text"."""
))
def func_gte(parser, x, y, _type=None):
    return _cmp(operator.ge, x, y, _type)


@script_function(documentation=N_(
    """`$len(text)`

Returns the number of characters in `text`."""
))
def func_len(parser, text=""):
    return str(len(text))


@script_function(eval_args=False, documentation=N_(
    """`$lenmulti(name,separator="; ")`

Returns the number of elements in the multi-value tag `name`. A literal value
    representing a multi-value can be substituted for `name`, using the
    `separator` (or "; " if not passed) to coerce the value into a proper
    multi-valued tag.

Example:

    $lenmulti(One; Two; Three) = 3
"""
))
def func_lenmulti(parser, multi, separator=MULTI_VALUED_JOINER):
    return str(len(MultiValue(parser, multi, separator)))


@script_function(documentation=N_(
    """`$performer(pattern="",join=", ")`

Returns the performers where the performance type (e.g. "vocal") matches `pattern`, joined by `join`.
You can specify a regular expression in the format `/pattern/flags`. `flags` are optional. Currently
the only supported flag is "i" (ignore case). For example `$performer(/^guitars?$/i)` matches the
performance type "guitar" or "Guitars", but not e.g. "bass guitar".

_Since Picard 0.10_"""
))
def func_performer(parser, pattern="", join=", "):
    values = []
    try:
        regex = pattern_as_regex(pattern, allow_wildcards=False)
    except re.error:
        return ''
    for name, value in parser.context.items():
        if name.startswith("performer:"):
            name, performance = name.split(':', 1)
            if regex.search(performance):
                values.append(value)
    return join.join(values)


@script_function(eval_args=False, documentation=N_(
    """`$matchedtracks()`

Returns the number of matched tracks within a release.
    **Only works in File Naming scripts.**

_Since Picard 0.12_"""
))
def func_matchedtracks(parser, *args):
    # only works in file naming scripts, always returns zero in tagging scripts
    file = parser.file
    if file and file.parent and hasattr(file.parent, 'album') and file.parent.album:
        return str(parser.file.parent.album.get_num_matched_tracks())
    return '0'


@script_function(documentation=N_(
    """`$is_complete()`

Returns true if every track in the album is matched to a single file.
**Only works in File Naming scripts.**"""
))
def func_is_complete(parser):
    # only works in file naming scripts, always returns zero in tagging scripts
    file = parser.file
    if (file and file.parent and hasattr(file.parent, 'album') and file.parent.album
            and file.parent.album.is_complete()):
        return '1'
    return ''


@script_function(documentation=N_(
    """`$firstalphachar(text,nonalpha="#")`

Returns the first character of `text`.
If `text` does not begin with an alphabetic character, then `nonalpha` is returned instead. If `nonalpha` is not
specified, the default value "#" will be used.

_Since Picard 0.12_"""
))
def func_firstalphachar(parser, text="", nonalpha="#"):
    if len(text) == 0:
        return nonalpha
    firstchar = text[0]
    if firstchar.isalpha():
        return firstchar.upper()
    else:
        return nonalpha


@script_function(documentation=N_(
    """`$initials(text)`

Returns the first character of each word in `text`, if it is an alphabetic character.

_Since Picard 0.12_"""
))
def func_initials(parser, text=""):
    return ''.join(a[:1] for a in text.split(" ") if a[:1].isalpha())


@script_function(documentation=N_(
    """`$firstwords(text,length)`

Like `$truncate()` except that it will only return the complete words from `text` which fit within `length` characters.

_Since Picard 0.12_"""
))
def func_firstwords(parser, text, length):
    try:
        length = int(length)
    except ValueError:
        length = 0
    if len(text) <= length:
        return text
    else:
        try:
            if text[length] == ' ':
                return text[:length]
            return text[:length].rsplit(' ', 1)[0]
        except IndexError:
            return ''


@script_function(documentation=N_(
    """`$startswith(text,prefix)`

Returns true if `text` starts with `prefix`.

_Since Picard 1.4_"""
))
def func_startswith(parser, text, prefix):
    if text.startswith(prefix):
        return '1'
    return ''


@script_function(documentation=N_(
    """`$endswith(text,suffix)`

Returns true if `text` ends with `suffix`.

_Since Picard 1.4_"""
))
def func_endswith(parser, text, suffix):
    if text.endswith(suffix):
        return '1'
    return ''


@script_function(documentation=N_(
    """`$truncate(text,length)`

Truncate `text` to `length`.

_Since Picard 0.12_"""
))
def func_truncate(parser, text, length):
    try:
        length = int(length)
    except ValueError:
        length = None
    return text[:length].rstrip()


@script_function(check_argcount=False, documentation=N_(
    """`$swapprefix(text,prefix1,prefix2,…)`

Moves the specified prefixes from the beginning to the end of `text`. Multiple
prefixes can be specified as separate parameters. If no prefix is specified 'A'
and 'The' are used by default.

Example:

    $swapprefix(%albumartist%,A,An,The,La,Le,Les,Un,Une)

_Since Picard 1.3, previously as a plugin since Picard 0.13_"""
))
def func_swapprefix(parser, text, *prefixes):
    # Inspired by the swapprefix plugin by Philipp Wolfer.

    text, prefix = _delete_prefix(parser, text, *prefixes)
    if prefix != '':
        return text + ', ' + prefix
    return text


@script_function(check_argcount=False, documentation=N_(
    """`$delprefix(text,prefix1,prefix2,…)`

Deletes the specified prefixes from the beginning of `text`. Multiple
prefixes can be specified as separate parameters.  If no prefix is specified 'A'
and 'The' are used by default.

Example:

    $delprefix(%albumartist%,A,An,The,La,Le,Les,Un,Une)

_Since Picard 1.3_"""
))
def func_delprefix(parser, text, *prefixes):
    # Inspired by the swapprefix plugin by Philipp Wolfer.

    return _delete_prefix(parser, text, *prefixes)[0]


def _delete_prefix(parser, text, *prefixes):
    """
    Worker function to deletes the specified prefixes.
    Returns remaining string and deleted part separately.
    If no prefix is specified 'A' and 'The' used.
    """
    # Inspired by the swapprefix plugin by Philipp Wolfer.

    if not prefixes:
        prefixes = ('A', 'The')
    text = text.strip()
    rx = '(' + r'\s+)|('.join(map(re.escape, prefixes)) + r'\s+)'
    match = re.match(rx, text)
    if match:
        pref = match.group()
        return text[len(pref):], pref.strip()
    return text, ''


@script_function(check_argcount=False, documentation=N_(
    """`$eq_any(x,a1,a2,…)`

Returns true if `x` equals `a1` or `a2` or …
Functionally equivalent to `$or($eq(x,a1),$eq(x,a2),…)`.
Functionally equivalent to the eq2 plugin."""
))
def func_eq_any(parser, x, *args):
    # Inspired by the eq2 plugin by Brian Schweitzer.
    return '1' if x in args else ''


@script_function(check_argcount=False, documentation=N_(
    """`$ne_all(x,a1,a2,…)`

Returns true if `x` does not equal `a1` and `a2` and …
Functionally equivalent to `$and($ne(x,a1),$ne(x,a2),…)`.
Functionally equivalent to the ne2 plugin."""
))
def func_ne_all(parser, x, *args):
    # Inspired by the ne2 plugin by Brian Schweitzer.
    return '1' if x not in args else ''


@script_function(check_argcount=False, documentation=N_(
    """`$eq_all(x,a1,a2,…)`

Returns true if `x` equals `a1` and `a2` and …
Functionally equivalent to `$and($eq(x,a1),$eq(x,a2),…)`.

Example:

    $if($eq_all(%albumartist%,%artist%,Justin Bieber),$set(engineer,Meat Loaf))
"""
))
def func_eq_all(parser, x, *args):
    for i in args:
        if x != i:
            return ''
    return '1'


@script_function(check_argcount=False, documentation=N_(
    """`$ne_any(x,a1,a2,…)`

Returns true if `x` does not equal `a1` or `a2` or …
Functionally equivalent to `$or($ne(x,a1),$ne(x,a2),…)`.

Example:

    $if($ne_any(%albumartist%,%trackartist%,%composer%),$set(lyricist,%composer%))
"""
))
def func_ne_any(parser, x, *args):
    return func_not(parser, func_eq_all(parser, x, *args))


@script_function(documentation=N_(
    """`$title(text)`

Returns `text` in title case (first character in every word capitalized).

Example:

    $set(album,$title(%album%))
_Since Picard 2.1_"""
))
def func_title(parser, text):
    return titlecase(text)


@script_function(documentation=N_(
    """`$is_audio()`

Returns true, if the file processed is an audio file.

_Since Picard 2.2_"""
))
def func_is_audio(parser):
    if func_is_video(parser) == "1":
        return ''
    else:
        return '1'


@script_function(documentation=N_(
    """`$is_video()`

Returns true, if the file processed is an video file.

_Since Picard 2.2_"""
))
def func_is_video(parser):
    if parser.context['~video'] and parser.context['~video'] != '0':
        return '1'
    else:
        return ''


@script_function(documentation=N_(
    """`$find(haystack,needle)`

Finds the location of one string within another.
    Returns the index of the first occurrence of `needle` in `haystack`, or "" if `needle` was not found.

_Since Picard 2.3_

Note that prior to Picard 2.3.2 `$find` returned "-1" if `needle` was not found."""
))
def func_find(parser, haystack, needle):
    index = haystack.find(needle)
    if index < 0:
        return ''
    return str(index)


@script_function(documentation=N_(
    """`$reverse(text)`

Returns `text` in reverse order."""
))
def func_reverse(parser, text):
    return text[::-1]


@script_function(documentation=N_(
    """`$substr(text,start[,end])`

Returns the substring beginning with the character at the `start` index, up to
    (but not including) the character at the `end` index. Indexes are
    zero-based. Negative numbers will be counted back from the end of the
    string. If the `start` or `end` indexes are left blank, they will default to
    the start and end of the string respectively."""
))
def func_substr(parser, text, start_index, end_index=None):
    try:
        start = int(start_index) if start_index else None
    except ValueError:
        start = None
    try:
        end = int(end_index) if end_index else None
    except ValueError:
        end = None
    return text[start:end]


@script_function(eval_args=False, documentation=N_(
    """`$getmulti(name,index,separator="; ")`

Gets the element at `index` from the multi-value tag `name`. A literal value
    representing a multi-value can be substituted for `name`, using the
    separator (or "; " if not passed) to coerce the value into a proper
    multi-valued tag."""
))
def func_getmulti(parser, multi, item_index, separator=MULTI_VALUED_JOINER):
    if not item_index:
        return ''
    try:
        index = int(item_index.eval(parser))
        multi_value = MultiValue(parser, multi, separator)
        return str(multi_value[index])
    except (ValueError, IndexError):
        return ''


@script_function(eval_args=False, documentation=N_(
    """`$foreach(name,code,separator="; ")`

Iterates over each element found in the multi-value tag `name`, executing
    `code`. For each loop, the element value is first stored in the tag
    `_loop_value` and the count is stored in the tag `_loop_count`. This allows
    the element or count value to be accessed within the `code` script. A
    literal value representing a multi-value can be substituted for `name`,
    using the separator (or "; " if not passed) to coerce the value into a
    proper multi-valued tag."""
))
def func_foreach(parser, multi, loop_code, separator=MULTI_VALUED_JOINER):
    multi_value = MultiValue(parser, multi, separator)
    for loop_count, value in enumerate(multi_value, 1):
        func_set(parser, '_loop_count', str(loop_count))
        func_set(parser, '_loop_value', str(value))
        loop_code.eval(parser)
    func_unset(parser, '_loop_count')
    func_unset(parser, '_loop_value')
    return ''


@script_function(eval_args=False, documentation=N_(
    """`$while(condition,code)`

Standard 'while' loop. Executes `code` repeatedly until `condition` no longer
    evaluates to `True`. For each loop, the count is stored in the tag
    `_loop_count`. This allows the count value to be accessed within the `code`
    script. The function limits the maximum number of iterations to 1000 as a
    safeguard against accidentally creating an infinite loop."""
))
def func_while(parser, condition, loop_code):
    if condition and loop_code:
        runaway_check = 1000
        loop_count = 0
        while condition.eval(parser) and loop_count < runaway_check:
            loop_count += 1
            func_set(parser, '_loop_count', str(loop_count))
            loop_code.eval(parser)
        func_unset(parser, '_loop_count')
    return ''


@script_function(eval_args=False, documentation=N_(
    """`$map(name,code,separator="; ")`

Iterates over each element found in the multi-value tag `name` and updates the
    value of the element to the value returned by `code`, returning the updated
    multi-value tag. For each loop, the element value is first stored in the tag
    `_loop_value` and the count is stored in the tag `_loop_count`. This allows
    the element or count value to be accessed within the `code` script.

Empty elements are automatically removed.

Example:

    $map(First:A; Second:B,$upper(%_loop_count%=%_loop_value%))

Result: 1=FIRST:A; 2=SECOND:B
"""
))
def func_map(parser, multi, loop_code, separator=MULTI_VALUED_JOINER):
    multi_value = MultiValue(parser, multi, separator)
    for loop_count, value in enumerate(multi_value, 1):
        func_set(parser, '_loop_count', str(loop_count))
        func_set(parser, '_loop_value', str(value))
        # Make changes in-place
        multi_value[loop_count - 1] = str(loop_code.eval(parser))
    func_unset(parser, '_loop_count')
    func_unset(parser, '_loop_value')
    return str(multi_value)


@script_function(eval_args=False, documentation=N_(
    """`$join(name,text,separator="; ")`

Joins all elements in `name`, placing `text` between each element, and returns the result as a string."""
))
def func_join(parser, multi, join_phrase, separator=MULTI_VALUED_JOINER):
    join_phrase = str(join_phrase.eval(parser))
    multi_value = MultiValue(parser, multi, separator)
    return join_phrase.join(multi_value)


@script_function(eval_args=False, documentation=N_(
    """`$slice(name,start,end,separator="; ")`

Returns a multi-value variable containing the elements between the `start` and
    `end` indexes from the multi-value tag `name`. A literal value representing
    a multi-value can be substituted for `name`, using the separator (or "; " if
    not passed) to coerce the value into a proper multi-valued tag. Indexes are
    zero based. Negative numbers will be counted back from the end of the list.
    If the `start` or `end` indexes are left blank, they will default to the
    start and end of the list respectively.

The following example will create a multi-value variable with all artists
    in `%artists%` except the first, which can be used to create a "feat." list.

Examples:

    $setmulti(supporting_artists,$slice(%artists%,1))
    $setmulti(supporting_artists,$slice(%artists%,1,-1))
"""
))
def func_slice(parser, multi, start_index, end_index=None, separator=MULTI_VALUED_JOINER):
    try:
        start = int(start_index.eval(parser)) if start_index else None
    except ValueError:
        start = None
    try:
        end = int(end_index.eval(parser)) if end_index else None
    except ValueError:
        end = None
    multi_value = MultiValue(parser, multi, separator)
    return multi_value.separator.join(multi_value[start:end])


@script_function(documentation=N_(
    """`$datetime(format="%Y-%m-%d %H:%M:%S")`

Returns the current date and time in the specified `format`, which is based on
    the standard Python `strftime` [format codes](https://strftime.org/). If no
    `format` is specified the date/time will be returned in the form
    `2020-02-05 14:26:32`.
Note: Platform-specific formatting codes should be avoided to help ensure the
    portability of scripts across the different platforms.  These codes include:
    remove zero-padding (e.g. `%-d` and `%-m` on Linux or macOS, and their
    equivalent `%#d` and `%#m` on Windows); element length specifiers (e.g.
    `%3Y`); and hanging '%' at the end of the format string."""
))
def func_datetime(parser, format=None):
    # local_tz required for Python 3.5 which does not allow setting astimezone()
    # on a naive datetime.datetime object.  This provides timezone information to
    # allow the use of %Z and %z in the output format.
    local_tz = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo

    # Handle case where format evaluates to ''
    if not format:
        format = '%Y-%m-%d %H:%M:%S'
    try:
        return datetime.datetime.now(tz=local_tz).strftime(format)
    except ValueError:
        stackitem = parser._function_stack.get()
        raise ScriptRuntimeError(stackitem, "Unsupported format code")


@script_function(eval_args=False, documentation=N_(
    """`$sortmulti(name,separator="; ")`

Returns a copy of the multi-value tag `name` with the elements sorted in ascending order.

Example:

    $sortmulti(B; A; C)

Result: A; B; C
"""
))
def func_sortmulti(parser, multi, separator=MULTI_VALUED_JOINER):
    multi_value = MultiValue(parser, multi, separator)
    return multi_value.separator.join(sorted(multi_value))


@script_function(eval_args=False, documentation=N_(
    """`$reversemulti(name,separator="; ")`

Returns a copy of the multi-value tag `name` with the elements in reverse order.
    This can be used in conjunction with the `$sortmulti` function to sort in
    descending order.

Example:

    $reversemulti($sortmulti(B; A; C))

Result: C; B; A
"""
))
def func_reversemulti(parser, multi, separator=MULTI_VALUED_JOINER):
    multi_value = MultiValue(parser, multi, separator)
    return multi_value.separator.join(reversed(multi_value))


@script_function(eval_args=False, documentation=N_(
    """`$unique(name,case_sensitive="",separator="; ")`

Returns a copy of the multi-value tag `name` with no duplicate elements.
    By default, a case-insensitive comparison of the elements is performed.

Example 1:

    $setmulti(foo,a; A; B; b; cd; Cd; cD; CD; a; A; b)
    $unique(%foo%)

Result: A; CD; b

Example 2:

    $setmulti(foo,a; A; B; b; a; b; A; B, cd)
    $unique(%foo%,True)

Result: A; B; a; b; cd
"""
))
def func_unique(parser, multi, case_sensitive="", separator=MULTI_VALUED_JOINER):
    multi_value = MultiValue(parser, multi, separator)
    if not case_sensitive:
        multi_value._multi = list({v.lower(): v for v in multi_value}.values())
    return multi_value.separator.join(sorted(set(multi_value)))


@script_function(documentation=N_(
    """`$countryname(country_code,translate="")`

Returns the name of the country for the specified country code.  If the country code is invalid an empty string will be returned.
If translate is not blank, the output will be translated into the current locale language.
"""
))
def func_countryname(parser, country_code, translate=""):
    name = RELEASE_COUNTRIES.get(country_code.strip().upper(), "")
    if translate:
        return gettext_countries(name)
    return name


DateTuple = namedtuple('DateTuple', ('year', 'month', 'day'))


def _split_date(date_to_parse, date_order="ymd"):
    """Split the specified date into parts.

    Args:
        date_to_parse (str): Date string to parse
        date_order (str, optional): Order of date elements. Can be "ymd", "mdy" or "dmy". Defaults to "ymd".

    Returns:
        tuple: Tuple of the date parts as (year, month, day)
    """

    parts = re.split(r'\D+', date_to_parse.strip())
    parts.extend(['', '', ''])
    date_order = date_order.lower()
    if date_order == 'dmy':
        return DateTuple(parts[2], parts[1], parts[0])
    elif date_order == 'mdy':
        return DateTuple(parts[2], parts[0], parts[1])
    else:
        return DateTuple(parts[0], parts[1], parts[2])


@script_function(documentation=N_(
    """`$year(date,date_order="ymd")`

Returns the year portion of the specified date.  The default order is "ymd".  This can be changed by specifying
either "dmy" or "mdy".  If the date is invalid an empty string will be returned.

_Since Picard 2.7_"""
))
def func_year(parser, date_to_parse, date_order='ymd'):
    return _split_date(date_to_parse, date_order).year


@script_function(documentation=N_(
    """`$month(date,date_order="ymd")`

Returns the month portion of the specified date.  The default order is "ymd".  This can be changed by specifying
either "dmy" or "mdy".  If the date is invalid an empty string will be returned.

_Since Picard 2.7_"""
))
def func_month(parser, date_to_parse, date_order='ymd'):
    return _split_date(date_to_parse, date_order).month


@script_function(documentation=N_(
    """`$day(date,date_order="ymd")`

Returns the day portion of the specified date.  The default order is "ymd".  This can be changed by specifying
either "dmy" or "mdy".  If the date is invalid an empty string will be returned.

_Since Picard 2.7_"""
))
def func_day(parser, date_to_parse, date_order='ymd'):
    return _split_date(date_to_parse, date_order).day


@script_function(documentation=N_(
    """`$dateformat(date,format="%Y-%m-%d",date_order="ymd")`

Returns the input date in the specified `format`, which is based on the standard
    Python `strftime` [format codes](https://strftime.org/). If no `format` is
    specified the date will be returned in the form `2020-02-05`.  If the date or
    format are invalid an empty string will be returned.

    The default order for the input date is "ymd".  This can be changed by specifying
    either "dmy" or "mdy".
Note: Platform-specific formatting codes should be avoided to help ensure the
    portability of scripts across the different platforms.  These codes include:
    remove zero-padding (e.g. `%-d` and `%-m` on Linux or macOS, and their
    equivalent `%#d` and `%#m` on Windows); element length specifiers (e.g.
    `%3Y`); and hanging '%' at the end of the format string.

_Since Picard 2.7_"""
))
def func_dateformat(parser, date_to_parse, date_format=None, date_order='ymd'):
    # Handle case where format evaluates to ''
    if not date_format:
        date_format = '%Y-%m-%d'
    yr, mo, da = _split_date(date_to_parse, date_order)
    try:
        date_object = datetime.date(int(yr), int(mo), int(da))
    except ValueError:
        return ''
    try:
        return date_object.strftime(date_format)
    except ValueError:
        return ''


@script_function(eval_args=False, documentation=N_(
    """`$is_multi(name)`

Returns '1' if the argument is a multi-value tag and there are more than one elements, otherwise an empty string.

Example:

    $is_multi(%artists%)

Result: 1 if there is more than one artist, otherwise "".

_Since Picard 2.7_"""
))
def func_is_multi(parser, multi):
    multi_value = MultiValue(parser, multi, MULTI_VALUED_JOINER)
    return '' if len(multi_value) < 2 else '1'


@script_function(eval_args=True, documentation=N_(
    """`$cleanmulti(name)`

Removes all empty string elements from the multi-value variable.

Example:

    $setmulti(test,one; ; two; three)
    $cleanmulti(test)

Result: Sets the value of 'test' to ["one", "two", "three"].

_Since Picard 2.8_"""
))
def func_cleanmulti(parser, multi):
    name = normalize_tagname(multi)
    values = [str(value) for value in parser.context.getall(name) if value or value == 0]
    parser.context[name] = values
    return ''


def _type_args(_type, *args):
    haystack = set()
    # Automatically expand multi-value arguments
    for item in args:
        haystack = haystack.union(set(x for x in item.split(MULTI_VALUED_JOINER)))
    if not _type:
        _type = 'auto'
    _typer = None
    if _type == 'auto':
        _type = 'text'
        for _test_type in (int, float):
            try:
                _type = set(_test_type(item) for item in haystack)
                _type = _test_type.__name__
                break
            except ValueError:
                pass
    _typer = None
    if _type == 'int':
        _typer = int
    elif _type == 'float':
        _typer = float
    elif _type in ('text', 'nocase'):
        pass
    else:
        # Unknown processing type
        raise ValueError
    if _typer is not None:
        haystack = set(_typer(item) for item in haystack)
    return haystack


def _extract(_func, _type, *args):
    try:
        haystack = _type_args(_type, *args)
    except ValueError:
        return ''

    if _type == 'nocase':
        op = operator.lt if _func == min else operator.gt
        val = None
        for item in haystack:
            if val is None or op(item.lower(), val.lower()):
                val = item
        return str(val)

    return str(_func(haystack))


@script_function(documentation=N_(
    """`$min(type,x,…)`

Returns the minimum value using the comparison specified in `type`.

Possible values of `type` are "int" (integer), "float" (floating point), "text"
(case-sensitive text), "nocase" (case-insensitive text) and "auto" (automatically
determine the type of arguments provided), with "auto" used as the default
comparison method if `type` is not specified.  The "auto" type will use the
first type that applies to both arguments in the following order of preference:
"int", "float" and "text".

Can be used with an arbitrary number of arguments.  Multi-value arguments
will be expanded automatically.

_Since Picard 2.9_"""
))
def func_min(parser, _type, x, *args):
    return _extract(min, _type, x, *args)


@script_function(documentation=N_(
    """`$max(type,x,…)`

Returns the maximum value using the comparison specified in `type`.

Possible values of `type` are "int" (integer), "float" (floating point), "text"
(case-sensitive text), "nocase" (case-insensitive text) and "auto" (automatically
determine the type of arguments provided), with "auto" used as the default
comparison method if `type` is not specified.  The "auto" type will use the
first type that applies to both arguments in the following order of preference:
"int", "float" and "text".

Can be used with an arbitrary number of arguments.  Multi-value arguments
will be expanded automatically.

_Since Picard 2.9_"""
))
def func_max(parser, _type, x, *args):
    return _extract(max, _type, x, *args)



# -*- coding: utf-8 -*-
#
# fix-header: nolicense
# MIT License
#
# Copyright(c) 2018 Konstantin Mochalov
# Copyright(c) 2022 Philipp Wolfer
# Copyright(c) 2022 Jeffrey Bosboom
#
# Original code from https://gist.github.com/kolen/765526
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re

from picard.disc.utils import (
    TocEntry,
    calculate_mb_toc_numbers,
)
from picard.util import detect_file_encoding


RE_TOC_TABLE_HEADER = re.compile(r""" \s*
    \s*.+\s+ \| # track
    \s+.+\s+ \| # start
    \s+.+\s+ \| # length
    \s+.+\s+ \| # start sector
    \s+.+\s*$   # end sector
    """, re.VERBOSE)

RE_TOC_TABLE_LINE = re.compile(r"""
    \s*
    (?P<num>\d+)
    \s*\|\s*
    (?P<start_time>[0-9:.]+)
    \s*\|\s*
    (?P<length_time>[0-9:.]+)
    \s*\|\s*
    (?P<start_sector>\d+)
    \s*\|\s*
    (?P<end_sector>\d+)
    \s*$""", re.VERBOSE)


def filter_toc_entries(lines):
    """
    Take iterator of lines, return iterator of toc entries
    """

    # Search the TOC table header
    for line in lines:
        # to allow internationalized EAC output where column headings
        # may differ
        if RE_TOC_TABLE_HEADER.match(line):
            # Skip over the table header separator
            next(lines)
            break

    for line in lines:
        m = RE_TOC_TABLE_LINE.search(line)
        if not m:
            break
        yield TocEntry(int(m['num']), int(m['start_sector']), int(m['end_sector']))


def toc_from_file(path):
    """Reads EAC / XLD / fre:ac log files, generates MusicBrainz disc TOC listing for use as discid.

    Warning: may work wrong for discs having data tracks. May generate wrong
    results on other non-standard cases."""
    encoding = detect_file_encoding(path)
    with open(path, 'r', encoding=encoding) as f:
        return calculate_mb_toc_numbers(filter_toc_entries(f))



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2019 Philipp Wolfer
# Copyright (C) 2020-2021 Laurent Monin
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


import os
import unittest

from picard import config
from picard.formats.ac3 import AC3File
from picard.formats.mutagenext.ac3 import native_ac3
from picard.metadata import Metadata

from .common import (
    CommonTests,
    load_metadata,
    save_and_load_metadata,
)
from .test_apev2 import CommonApeTests


class AC3WithAPETest(CommonApeTests.ApeTestCase):
    testfile = 'test.ac3'
    supports_ratings = False
    expected_info = {
        'length': 106,
        '~bitrate': '192.0',
        '~sample_rate': '44100',
        '~channels': '2',
    }
    unexpected_info = ['~video']

    def setUp(self):
        super().setUp()
        config.setting['ac3_save_ape'] = True
        config.setting['remove_ape_from_ac3'] = True

    @unittest.skipUnless(native_ac3, "mutagen.ac3 not available")
    def test_info(self):
        super().test_info()


class AC3NoTagsTest(CommonTests.BaseFileTestCase):
    testfile = 'test-apev2.ac3'

    def setUp(self):
        super().setUp()
        config.setting['ac3_save_ape'] = False
        config.setting['remove_ape_from_ac3'] = False

    def test_load_but_do_not_save_tags(self):
        metadata = load_metadata(self.filename)
        self.assertEqual('Test AC3 with APEv2 tags', metadata['title'])
        self.assertEqual('The Artist', metadata['artist'])
        metadata['artist'] = 'Foo'
        metadata['title'] = 'Bar'
        metadata = save_and_load_metadata(self.filename, metadata)
        self.assertEqual('Test AC3 with APEv2 tags', metadata['title'])
        self.assertEqual('The Artist', metadata['artist'])

    def test_remove_ape_tags(self):
        config.setting['remove_ape_from_ac3'] = True
        metadata = Metadata({
            'artist': 'Foo'
        })
        metadata = save_and_load_metadata(self.filename, metadata)
        self.assertEqual('AC-3', metadata['~format'])
        self.assertNotIn('title', metadata)
        self.assertNotIn('artist', metadata)

    def test_info_format(self):
        metadata = load_metadata(os.path.join('test', 'data', 'test.ac3'))
        self.assertEqual('AC-3', metadata['~format'])
        metadata = load_metadata(os.path.join('test', 'data', 'test-apev2.ac3'))
        self.assertEqual('AC-3 (APEv2)', metadata['~format'])
        if native_ac3:
            metadata = load_metadata(os.path.join('test', 'data', 'test.eac3'))
            self.assertEqual('Enhanced AC-3', metadata['~format'])

    def test_supports_tag(self):
        config.setting['ac3_save_ape'] = True
        self.assertTrue(AC3File.supports_tag('title'))
        config.setting['ac3_save_ape'] = False
        self.assertFalse(AC3File.supports_tag('title'))


@unittest.skipUnless(native_ac3, "mutagen.ac3 not available")
class EAC3Test(CommonTests.SimpleFormatsTestCase):
    testfile = 'test.eac3'
    expected_info = {
        '~format': 'Enhanced AC-3',
        'length': 107,
        '~sample_rate': '44100',
        '~channels': '2',
    }
    unexpected_info = ['~video']

    def setUp(self):
        super().setUp()
        config.setting['ac3_save_ape'] = True

    def test_bitrate(self):
        # For EAC3 bitrate is calculated and often a fractional value
        metadata = load_metadata(os.path.join('test', 'data', 'test.ac3'))
        self.assertAlmostEqual(192.0, float(metadata['~bitrate']))



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2022 skelly37
# Copyright (C) 2023 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import concurrent.futures
from platform import python_version
import time
import uuid

from test.picardtestcase import PicardTestCase

from picard.util import pipe


def pipe_listener(pipe_handler):
    while True:
        for message in pipe_handler.read_from_pipe():
            if message and message != pipe.Pipe.NO_RESPONSE_MESSAGE:
                return message


class TestPipe(PicardTestCase):
    # some random name that is different on each run
    NAME = str(uuid.uuid4())
    VERSION = python_version()

    def test_invalid_args(self):
        # Pipe should be able to make args iterable (last argument)
        self.assertRaises(pipe.PipeErrorInvalidArgs, pipe.Pipe, self.NAME, self.VERSION, 1)
        self.assertRaises(pipe.PipeErrorInvalidAppData, pipe.Pipe, 21, self.VERSION, None)
        self.assertRaises(pipe.PipeErrorInvalidAppData, pipe.Pipe, self.NAME, 21, None)

    def test_pipe_protocol(self):
        message = "foo"

        __pool = concurrent.futures.ThreadPoolExecutor()
        pipe_handler = pipe.Pipe(self.NAME, self.VERSION)
        try:
            plistener = __pool.submit(pipe_listener, pipe_handler)
            time.sleep(.2)
            res = ""

            # handle the write/read processes
            try:
                pipe_handler.send_to_pipe(message)
                res = plistener.result(timeout=6)
            except concurrent.futures._base.TimeoutError:
                pass

            self.assertEqual(res, message,
                            "Data is sent and read correctly")
        finally:
            time.sleep(.2)
            pipe_handler.stop()
            __pool.shutdown()



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2019-2021 Laurent Monin
# Copyright (C) 2020 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from test.picardtestcase import PicardTestCase

from picard import config
from picard.util.settingsoverride import SettingsOverride


class SettingsOverrideTest(PicardTestCase):

    def setUp(self):
        super().setUp()
        self.set_config_values({'key1': 'origval1', 'key2': 'origval2'})
        self.new_settings = {'key1': 'newval2'}

    def test_read_orig_settings(self):
        override = SettingsOverride(config.setting, self.new_settings)
        self.assertEqual(override['key1'], 'newval2')
        self.assertEqual(override['key2'], 'origval2')
        with self.assertRaises(KeyError):
            x = override['key3']  # noqa: F841

    def test_read_orig_settings_kw(self):
        override = SettingsOverride(config.setting, key1='newval2')
        self.assertEqual(override['key1'], 'newval2')
        self.assertEqual(override['key2'], 'origval2')

    def test_write_orig_settings(self):
        override = SettingsOverride(config.setting, self.new_settings)
        override['key1'] = 'newval3'
        self.assertEqual(override['key1'], 'newval3')
        self.assertEqual(config.setting['key1'], 'origval1')

        override['key2'] = 'newval4'
        self.assertEqual(override['key2'], 'newval4')
        self.assertEqual(config.setting['key2'], 'origval2')

        override['key3'] = 'newval5'
        self.assertEqual(override['key3'], 'newval5')
        with self.assertRaises(KeyError):
            x = config.setting['key3']  # noqa: F841

    def test_del_orig_settings(self):
        override = SettingsOverride(config.setting, self.new_settings)

        override['key1'] = 'newval3'
        self.assertEqual(override['key1'], 'newval3')
        del override['key1']
        self.assertEqual(override['key1'], 'origval1')

        self.assertEqual(override['key2'], 'origval2')
        del override['key2']
        self.assertEqual(override['key2'], 'origval2')



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2011 Lukáš Lalinský
# Copyright (C) 2017 Sambhav Kothari
# Copyright (C) 2018 Vishal Choudhary
# Copyright (C) 2018, 2020-2021 Laurent Monin
# Copyright (C) 2020, 2022 Philipp Wolfer
# Copyright (C) 2022 cybersphinx
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from functools import partial

from PyQt5 import QtCore

from picard import log
from picard.util import load_json


# Maximum difference between file duration and MB recording length.
# If the match is above this threshold the MBID will not get submitted.
# Compare also acoustid/const.py in acoustid-server sources
FINGERPRINT_MAX_ALLOWED_LENGTH_DIFF_MS = 30000


class Submission(object):

    def __init__(self, fingerprint, duration, recordingid=None, metadata=None):
        self.fingerprint = fingerprint
        self.duration = duration
        self.recordingid = self.orig_recordingid = recordingid
        self.metadata = metadata
        self.attempts = 0

    def __len__(self):
        # payload approximation
        # the length of all submitted data, plus a small overhead to account for
        # potential urlencode expansion
        return int(sum((len(key) + len(value) + 2 for key, value in self.args.items())) * 1.03)

    def __bool__(self):
        return bool(self.fingerprint and self.duration is not None)

    def __repr__(self):
        return '<%s %r, %d, %r>' % (self.__class__.__name__, self.fingerprint, self.duration, self.recordingid)

    @property
    def puid(self):
        return self.metadata.get('musicip_puid', '') if self.metadata else ''

    @property
    def valid_duration(self):
        return self.metadata is None or abs(self.duration * 1000 - self.metadata.length) <= FINGERPRINT_MAX_ALLOWED_LENGTH_DIFF_MS

    @property
    def is_submitted(self):
        return not self.recordingid or self.orig_recordingid == self.recordingid

    @property
    def args(self):
        """Returns a dictionary of arguments suitable for submission to AcoustID."""
        args = {
            'fingerprint': self.fingerprint,
            'duration': str(self.duration),
        }
        metadata = self.metadata
        if metadata:
            puid = metadata['musicip_puid']
            if puid:
                args['puid'] = puid
        if self.valid_duration and self.recordingid:
            args['mbid'] = self.recordingid
        elif metadata:
            args['track'] = metadata['title']
            args['artist'] = metadata['artist']
            args['album'] = metadata['album']
            args['albumartist'] = metadata['albumartist']
            args['trackno'] = metadata['tracknumber']
            args['discno'] = metadata['discnumber']
            year = metadata['year'] or metadata['date'][:4]
            if year and year.isdecimal():
                args['year'] = year
        return args


class AcoustIDManager(QtCore.QObject):

    # AcoustID has a post limit of around 1 MB.
    MAX_PAYLOAD = 1000000
    # Limit each submission to N attempts
    MAX_ATTEMPTS = 5
    # In case of Payload Too Large error, batch size is reduced by this factor
    # and a retry occurs
    BATCH_SIZE_REDUCTION_FACTOR = 0.7

    def __init__(self, acoustid_api):
        super().__init__()
        self._submissions = {}
        self._acoustid_api = acoustid_api

    def add(self, file, recordingid):
        if not file.acoustid_fingerprint or not file.acoustid_length:
            return
        metadata = file.metadata
        self._submissions[file] = Submission(
            file.acoustid_fingerprint, file.acoustid_length, recordingid, metadata)
        self._check_unsubmitted()

    def update(self, file, recordingid):
        submission = self._submissions.get(file)
        if submission is None:
            return
        submission.recordingid = recordingid
        submission.metadata = file.metadata
        self._check_unsubmitted()

    def remove(self, file):
        if file in self._submissions:
            del self._submissions[file]
        self._check_unsubmitted()

    def is_submitted(self, file):
        submission = self._submissions.get(file)
        if submission:
            return submission.is_submitted
        return True

    def _unsubmitted(self, reset=False):
        for file, submission in self._submissions.items():
            if not submission.is_submitted:
                if reset:
                    submission.attempts = 0
                yield (file, submission)

    def _check_unsubmitted(self):
        enabled = next(self._unsubmitted(), None) is not None
        self.tagger.window.enable_submit(enabled)

    def submit(self):
        self.max_batch_size = self.MAX_PAYLOAD
        submissions = list(self._unsubmitted(reset=True))
        if not submissions:
            self._check_unsubmitted()
            return
        log.debug("AcoustID: submitting total of %d fingerprints…", len(submissions))
        self._batch_submit(submissions)

    def _batch(self, submissions):
        batch = []
        remaining = []
        batch_size = 0
        max_attempts = self.MAX_ATTEMPTS
        for file, submission in submissions:
            if submission.attempts < max_attempts:
                batch_size += len(submission)
                if batch_size < self.max_batch_size:
                    submission.attempts += 1
                    batch.append((file, submission))
                    continue
                else:
                    # force appending the rest to remaining if we reach max_batch_size
                    max_attempts = 0
            remaining.append((file, submission))
        return batch, remaining

    def _batch_submit(self, submissions, errors=None):
        if not submissions:  # All fingerprints submitted, nothing to do
            if errors:
                log_msg = N_("AcoustID submission finished, but not all fingerprints have been submitted")
            else:
                log_msg = N_("AcoustID submission finished successfully")
            log.debug(log_msg)
            self.tagger.window.set_statusbar_message(
                log_msg, echo=None, timeout=3000)
            self._check_unsubmitted()
            return

        batch, submissions = self._batch(submissions)

        if not batch:
            if self.max_batch_size == 0:
                log_msg = N_("AcoustID submission failed permanently, maximum batch size reduced to zero")
            else:
                log_msg = N_("AcoustID submission failed permanently, probably too many retries")
            log.error(log_msg)
            self.tagger.window.set_statusbar_message(
                log_msg, echo=None, timeout=3000)
            self._check_unsubmitted()
            return

        log.debug("AcoustID: submitting batch of %d fingerprints (%d remaining)…",
            len(batch), len(submissions))
        self.tagger.window.set_statusbar_message(
            N_("Submitting AcoustIDs …"),
            echo=None
        )
        if not errors:
            errors = []
        self._acoustid_api.submit_acoustid_fingerprints(
            [submission for file_, submission in batch],
            partial(self._batch_submit_finished, submissions, batch, errors)
        )

    def _batch_submit_finished(self, submissions, batch, previous_errors, document, http, error):
        if error:
            # re-add batched items to remaining list
            submissions.extend(batch)

            response_code = self._acoustid_api.webservice.http_response_code(http)
            if response_code == 413:
                self.max_batch_size = int(self.max_batch_size * self.BATCH_SIZE_REDUCTION_FACTOR)
                log.warn("AcoustID: payload too large, batch size reduced to %d", self.max_batch_size)
            else:
                try:
                    errordoc = load_json(document)
                    message = errordoc['error']['message']
                except BaseException:
                    message = ""
                mparms = {
                    'error': http.errorString(),
                    'message': message
                }
                previous_errors.append(mparms)
                log_msg = N_("AcoustID submission failed with error '%(error)s': %(message)s")
                log.error(log_msg, mparms)
                self.tagger.window.set_statusbar_message(
                    log_msg, mparms, echo=None, timeout=3000)
        else:
            log.debug("AcoustID: %d fingerprints successfully submitted", len(batch))
            for file, submission in batch:
                submission.orig_recordingid = submission.recordingid
                file.update()
            self._check_unsubmitted()
        self._batch_submit(submissions, previous_errors)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2006-2007, 2011 Lukáš Lalinský
# Copyright (C) 2011-2013 Michael Wiencek
# Copyright (C) 2012 Chad Wilson
# Copyright (C) 2012-2013, 2018, 2021-2022, 2024 Philipp Wolfer
# Copyright (C) 2013, 2018, 2020-2021, 2024 Laurent Monin
# Copyright (C) 2016 Suhas
# Copyright (C) 2016-2017 Sambhav Kothari
# Copyright (C) 2018 Vishal Choudhary
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

from http.server import (
    BaseHTTPRequestHandler,
    HTTPServer,
)
import re
import threading
from urllib.parse import (
    parse_qs,
    urlparse,
)

from PyQt5 import QtCore

from picard import (
    PICARD_APP_NAME,
    PICARD_ORG_NAME,
    PICARD_VERSION_STR,
    log,
)
from picard.browser import addrelease
from picard.config import get_config
from picard.oauth import OAuthInvalidStateError
from picard.util import mbid_validate
from picard.util.thread import to_main


try:
    from http.server import ThreadingHTTPServer
except ImportError:
    from socketserver import ThreadingMixIn

    class ThreadingHTTPServer(ThreadingMixIn, HTTPServer):
        daemon_threads = True


SERVER_VERSION = '%s-%s/%s' % (PICARD_ORG_NAME, PICARD_APP_NAME, PICARD_VERSION_STR)
RE_VALID_ORIGINS = re.compile(r'^(?:[^\.]+\.)*musicbrainz\.org$')


def _is_valid_origin(origin):
    try:
        url = urlparse(origin)
    except ValueError:
        return False
    hostname = url.hostname
    if not hostname:
        return False
    if RE_VALID_ORIGINS.match(hostname):
        return True
    config = get_config()
    return config.setting['server_host'] == hostname


class BrowserIntegration(QtCore.QObject):

    listen_port_changed = QtCore.pyqtSignal(int)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.server = None

    @property
    def host_address(self):
        if not self.server:
            return ''
        return self.server.server_address[0]

    @property
    def port(self):
        if not self.server:
            return 0
        return self.server.server_address[1]

    @property
    def is_running(self):
        return self.server is not None

    def start(self):
        if self.server:
            self.stop()

        config = get_config()
        if config.setting["browser_integration_localhost_only"]:
            host_address = '127.0.0.1'
        else:
            host_address = '0.0.0.0'  # nosec

        try:
            for port in range(config.setting["browser_integration_port"], 65535):
                try:
                    self.server = ThreadingHTTPServer((host_address, port), RequestHandler)
                except OSError:
                    continue
                log.info("Starting the browser integration (%s:%d)", host_address, port)
                self.listen_port_changed.emit(port)
                threading.Thread(target=self.server.serve_forever).start()
                break
            else:
                log.error("Failed finding an available port for the browser integration.")
                self.stop()
        except Exception:
            log.error("Failed starting the browser integration on %s", host_address, exc_info=True)

    def stop(self):
        if self.server:
            try:
                log.info("Stopping the browser integration")
                self.server.shutdown()
                self.server.server_close()
                self.server = None
                self.listen_port_changed.emit(self.port)
            except Exception:
                log.error("Failed stopping the browser integration", exc_info=True)
        else:
            log.debug("Browser integration inactive, no need to stop")


class RequestHandler(BaseHTTPRequestHandler):

    def do_OPTIONS(self):
        origin = self.headers['origin']
        if _is_valid_origin(origin):
            self.send_response(204)
            self.send_header('Access-Control-Allow-Origin', origin)
            self.send_header('Access-Control-Allow-Methods', 'GET')
            self.send_header('Access-Control-Allow-Credentials', 'false')
            self.send_header('Access-Control-Allow-Private-Network', 'true')
            self.send_header('Access-Control-Max-Age', 3600)
            self.send_header('Vary', 'Origin')
        else:
            self.send_response(401)
        self.end_headers()

    def do_GET(self):
        try:
            self._handle_get()
        except Exception:
            log.error('Browser integration failed handling request', exc_info=True)
            self._response(500, 'Unexpected request error')

    def log_error(self, format, *args):
        log.error(format, *args)

    def log_message(self, format, *args):
        log.info(format, *args)

    def _handle_get(self):
        parsed = urlparse(self.path)
        args = parse_qs(parsed.query)
        action = parsed.path

        if action == '/':
            self._response(200, SERVER_VERSION)
        elif action == '/openalbum':
            self._load_mbid('album', args)
        elif action == '/opennat':
            self._load_mbid('nat', args)
        elif action == '/add' and addrelease.is_available():
            self._add_release(args)
        elif action == '/auth':
            self._auth(args)
        else:
            self._response(404, 'Unknown action.')

    def _load_mbid(self, type, args):
        if 'id' in args and args['id']:
            mbid = args['id'][0]
            if not mbid_validate(mbid):
                self._response(400, '"id" is not a valid MBID.')
            else:
                tagger = QtCore.QCoreApplication.instance()
                to_main(tagger.load_mbid, type, mbid)
                self._response(200, 'MBID "%s" loaded' % mbid)
        else:
            self._response(400, 'Missing parameter "id".')

    def _add_release(self, args):
        if 'token' in args and args['token']:
            try:
                content = addrelease.serve_form(args['token'][0])
                self._response(200, content, 'text/html')
            except addrelease.NotFoundError as err:
                self._response(404, str(err))
            except addrelease.InvalidTokenError:
                self._response(400, 'Invalid token')
        else:
            self._response(400, 'Missing parameter "token".')

    def _auth(self, args):
        if 'code' in args and args['code']:
            tagger = QtCore.QCoreApplication.instance()
            oauth_manager = tagger.webservice.oauth_manager
            try:
                state = args.get('state', [''])[0]
                callback = oauth_manager.verify_state(state)
            except OAuthInvalidStateError:
                self._response(400, 'Invalid "state" parameter.')
                return
            to_main(
                oauth_manager.exchange_authorization_code,
                authorization_code=args['code'][0],
                scopes='profile tag rating collection submit_isrc submit_barcode',
                callback=callback,
            )
            self._response(200, "Authentication successful, you can close this window now.", 'text/html')
        else:
            self._response(400, 'Missing parameter "code".')

    def _response(self, code, content='', content_type='text/plain'):
        self.server_version = SERVER_VERSION
        self.send_response(code)
        self.send_header('Content-Type', content_type)
        self.send_header('Cache-Control', 'max-age=0')
        origin = self.headers['origin']
        if _is_valid_origin(origin):
            self.send_header('Access-Control-Allow-Origin', origin)
            self.send_header('Vary', 'Origin')
        self.end_headers()
        self.wfile.write(content.encode())



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2020-2022 Philipp Wolfer
# Copyright (C) 2021 Laurent Monin
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


import unittest
from unittest.mock import (
    Mock,
    patch,
)

from test.picardtestcase import PicardTestCase

import picard.disc


test_toc = [1, 11, 242457, 150, 44942, 61305, 72755, 96360, 130485, 147315, 164275, 190702, 205412, 220437]


class MockDisc:

    id = 'lSOVc5h6IXSuzcamJS1Gp4_tRuA-'
    mcn = '5029343070452'
    tracks = list(range(0, 11))
    toc_string = ' '.join(str(i) for i in test_toc)
    submission_url = 'https://musicbrainz.org/cdtoc/attach?id=lSOVc5h6IXSuzcamJS1Gp4_tRuA-&tracks=11&toc=1+11+242457+150+44942+61305+72755+96360+130485+147315+164275+190702+205412+220437'


class DiscTest(PicardTestCase):
    @unittest.skipUnless(picard.disc.discid, "discid not available")
    def test_raise_disc_error(self):
        disc = picard.disc.Disc()
        self.assertRaises(picard.disc.discid.DiscError, disc.read, 'notadevice')

    def test_init_with_id(self):
        discid = 'theId'
        disc = picard.disc.Disc(id=discid)
        self.assertEqual(discid, disc.id)
        self.assertEqual(0, disc.tracks)
        self.assertIsNone(disc.toc_string)
        self.assertIsNone(disc.submission_url)

    @patch.object(picard.disc, 'discid')
    def test_read(self, mock_discid):
        self.set_config_values(setting={
            'server_host': 'musicbrainz.org',
            'server_port': 443,
            'use_server_for_submission': False,
        })
        mock_discid.read = Mock(return_value=MockDisc())
        device = '/dev/cdrom1'
        disc = picard.disc.Disc()
        self.assertEqual(None, disc.id)
        self.assertEqual(None, disc.mcn)
        self.assertEqual(0, disc.tracks)
        self.assertEqual(None, disc.toc_string)
        self.assertEqual(None, disc.submission_url)
        disc.read(device)
        mock_discid.read.assert_called_with(device, features=['mcn'])
        self.assertEqual(MockDisc.id, disc.id)
        self.assertEqual(MockDisc.mcn, disc.mcn)
        self.assertEqual(11, disc.tracks)
        self.assertEqual(MockDisc.toc_string, disc.toc_string)
        self.assertEqual(MockDisc.submission_url, disc.submission_url)

    @patch.object(picard.disc, 'discid')
    def test_put(self, mock_discid):
        mock_discid.put = Mock(return_value=MockDisc())
        disc = picard.disc.Disc()
        disc.put(test_toc)
        self.assertEqual(MockDisc.id, disc.id)

    @patch.object(picard.disc, 'discid')
    def test_put_invalid_toc_1(self, mock_discid):
        mock_discid.TOCError = Exception
        disc = picard.disc.Disc()
        with self.assertRaises(mock_discid.TOCError):
            disc.put([1, 11])

    @patch.object(picard.disc, 'discid')
    def test_put_invalid_toc_2(self, mock_discid):
        mock_discid.TOCError = Exception
        mock_discid.put = Mock(side_effect=mock_discid.TOCError)
        disc = picard.disc.Disc()
        with self.assertRaises(mock_discid.TOCError):
            disc.put([1, 11, 242457])

    @patch.object(picard.disc, 'discid')
    def test_submission_url(self, mock_discid):
        self.set_config_values(setting={
            'server_host': 'test.musicbrainz.org',
            'server_port': 80,
            'use_server_for_submission': True,
        })
        mock_discid.read = Mock(return_value=MockDisc())
        disc = picard.disc.Disc()
        disc.read()
        self.assertEqual(
            'http://test.musicbrainz.org/cdtoc/attach?id=lSOVc5h6IXSuzcamJS1Gp4_tRuA-&tracks=11&toc=1+11+242457+150+44942+61305+72755+96360+130485+147315+164275+190702+205412+220437',
            disc.submission_url
        )



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2020 Philipp Wolfer
# Copyright (C) 2020-2021 Laurent Monin
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from test.picardtestcase import PicardTestCase

from picard import config
from picard.util.preservedtags import PreservedTags


class PreservedTagsTest(PicardTestCase):
    def setUp(self):
        super().setUp()
        config.setting[PreservedTags.opt_name] = ["tag1", "tag2"]

    def test_load_and_contains(self):
        preserved = PreservedTags()
        self.assertIn("tag1", preserved)
        self.assertIn("tag2", preserved)
        self.assertIn("TAG1", preserved)
        self.assertIn(" tag1", preserved)

    def test_add(self):
        preserved = PreservedTags()
        self.assertNotIn("tag3", preserved)
        preserved.add("tag3")
        self.assertIn("tag3", preserved)
        # Add must persists the change
        self.assertIn("tag3", PreservedTags())

    def test_add_case_insensitive(self):
        preserved = PreservedTags()
        self.assertNotIn("tag3", preserved)
        preserved.add("TAG3")
        self.assertIn("tag3", preserved)

    def test_discard(self):
        preserved = PreservedTags()
        self.assertIn("tag1", preserved)
        preserved.discard("tag1")
        self.assertNotIn("tag1", preserved)
        # Discard must persists the change
        self.assertNotIn("tag1", PreservedTags())

    def test_discard_case_insensitive(self):
        preserved = PreservedTags()
        self.assertIn("tag1", preserved)
        preserved.discard("TAG1")
        self.assertNotIn("tag1", preserved)

    def test_order(self):
        preserved = PreservedTags()
        preserved.add('tag3')
        preserved.add('tag2')
        preserved.add('tag1')
        preserved.discard('tag2')
        self.assertEqual(config.setting[PreservedTags.opt_name], ['tag1', 'tag3'])



# -*- coding: utf-8 -*-

# Automatically generated - don't edit.
# Use `python setup.py build_ui` to update it.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(800, 450)
        self.vboxlayout = QtWidgets.QVBoxLayout(Dialog)
        self.vboxlayout.setContentsMargins(9, 9, 9, 9)
        self.vboxlayout.setSpacing(6)
        self.vboxlayout.setObjectName("vboxlayout")
        self.dialog_splitter = QtWidgets.QSplitter(Dialog)
        self.dialog_splitter.setOrientation(QtCore.Qt.Horizontal)
        self.dialog_splitter.setChildrenCollapsible(False)
        self.dialog_splitter.setObjectName("dialog_splitter")
        self.pages_tree = QtWidgets.QTreeWidget(self.dialog_splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pages_tree.sizePolicy().hasHeightForWidth())
        self.pages_tree.setSizePolicy(sizePolicy)
        self.pages_tree.setMinimumSize(QtCore.QSize(140, 0))
        self.pages_tree.setObjectName("pages_tree")
        self.pages_stack = QtWidgets.QStackedWidget(self.dialog_splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pages_stack.sizePolicy().hasHeightForWidth())
        self.pages_stack.setSizePolicy(sizePolicy)
        self.pages_stack.setMinimumSize(QtCore.QSize(280, 0))
        self.pages_stack.setObjectName("pages_stack")
        self.vboxlayout.addWidget(self.dialog_splitter)
        self.buttonbox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonbox.setMinimumSize(QtCore.QSize(0, 0))
        self.buttonbox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonbox.setObjectName("buttonbox")
        self.vboxlayout.addWidget(self.buttonbox)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_("Options"))



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2016 Rahul Raturi
# Copyright (C) 2018-2022 Laurent Monin
# Copyright (C) 2018-2023 Philipp Wolfer
# Copyright (C) 2020 Ray Bouchard
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from collections import namedtuple

from PyQt5 import (
    QtCore,
    QtGui,
    QtNetwork,
    QtWidgets,
)

from picard.config import get_config
from picard.util import (
    icontheme,
    restore_method,
)

from picard.ui.tablebaseddialog import TableBasedDialog
from picard.ui.util import StandardButton


class SearchBox(QtWidgets.QWidget):

    def __init__(self, parent, force_advanced_search=None):
        super().__init__(parent)
        self.search_action = QtWidgets.QAction(icontheme.lookup('system-search'), _("Search"), self)
        self.search_action.setEnabled(False)
        self.search_action.triggered.connect(self.search)
        if force_advanced_search is None:
            config = get_config()
            self.force_advanced_search = False
            self.use_advanced_search = config.setting['use_adv_search_syntax']
        else:
            self.force_advanced_search = True
            self.use_advanced_search = force_advanced_search
        self.setupUi()

    def focus_in_event(self, event):
        # When focus is on search edit box, need to disable
        # dialog's accept button. This would avoid closing of dialog when user
        # hits enter.
        parent = self.parent()
        if parent.table:
            parent.table.clearSelection()
        parent.accept_button.setEnabled(False)

    def setupUi(self):
        self.layout = QtWidgets.QVBoxLayout(self)
        self.search_row_widget = QtWidgets.QWidget(self)
        self.search_row_layout = QtWidgets.QHBoxLayout(self.search_row_widget)
        self.search_row_layout.setContentsMargins(1, 1, 1, 1)
        self.search_row_layout.setSpacing(1)
        self.search_edit = QtWidgets.QLineEdit(self.search_row_widget)
        self.search_edit.setClearButtonEnabled(True)
        self.search_edit.returnPressed.connect(self.trigger_search_action)
        self.search_edit.textChanged.connect(self.enable_search)
        self.search_edit.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.search_edit.focusInEvent = self.focus_in_event
        self.search_row_layout.addWidget(self.search_edit)
        self.search_button = QtWidgets.QToolButton(self.search_row_widget)
        self.search_button.setAutoRaise(True)
        self.search_button.setDefaultAction(self.search_action)
        self.search_button.setIconSize(QtCore.QSize(22, 22))
        self.search_row_layout.addWidget(self.search_button)
        self.search_row_widget.setLayout(self.search_row_layout)
        self.layout.addWidget(self.search_row_widget)
        self.adv_opt_row_widget = QtWidgets.QWidget(self)
        self.adv_opt_row_layout = QtWidgets.QHBoxLayout(self.adv_opt_row_widget)
        self.adv_opt_row_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
        self.adv_opt_row_layout.setContentsMargins(1, 1, 1, 1)
        self.adv_opt_row_layout.setSpacing(1)
        self.use_adv_search_syntax = QtWidgets.QCheckBox(self.adv_opt_row_widget)
        self.use_adv_search_syntax.setText(_("Use advanced query syntax"))
        self.use_adv_search_syntax.setChecked(self.use_advanced_search)
        self.use_adv_search_syntax.stateChanged.connect(self.update_advanced_syntax_setting)
        self.adv_opt_row_layout.addWidget(self.use_adv_search_syntax)
        self.adv_syntax_help = QtWidgets.QLabel(self.adv_opt_row_widget)
        self.adv_syntax_help.setOpenExternalLinks(True)
        self.adv_syntax_help.setText(_(
            "&#160;(<a href='https://musicbrainz.org/doc/Indexed_Search_Syntax'>"
            "Syntax Help</a>)"))
        self.adv_opt_row_layout.addWidget(self.adv_syntax_help)
        self.adv_opt_row_widget.setLayout(self.adv_opt_row_layout)
        self.layout.addWidget(self.adv_opt_row_widget)
        self.layout.setContentsMargins(1, 1, 1, 1)
        self.layout.setSpacing(1)
        self.setMaximumHeight(60)

    def search(self):
        self.parent().search(self.query)

    def restore_checkbox_state(self):
        self.use_adv_search_syntax.setChecked(self.use_advanced_search)

    def update_advanced_syntax_setting(self):
        self.use_advanced_search = self.use_adv_search_syntax.isChecked()
        if not self.force_advanced_search:
            config = get_config()
            config.setting['use_adv_search_syntax'] = self.use_advanced_search

    def enable_search(self):
        if self.query:
            self.search_action.setEnabled(True)
        else:
            self.search_action.setEnabled(False)

    def trigger_search_action(self):
        if self.search_action.isEnabled():
            self.search_action.trigger()

    def get_query(self):
        return self.search_edit.text()

    def set_query(self, query):
        return self.search_edit.setText(query)

    query = property(get_query, set_query)


Retry = namedtuple('Retry', ['function', 'query'])


class SearchDialog(TableBasedDialog):
    accept_button_title = ""

    def __init__(self, parent, accept_button_title, show_search=True, search_type=None, force_advanced_search=None):
        self.accept_button_title = accept_button_title
        self.search_results = []
        self.show_search = show_search
        self.search_type = search_type
        self.force_advanced_search = force_advanced_search
        self.search_box = None
        super().__init__(parent)

    @property
    def use_advanced_search(self):
        if self.show_search:
            return self.search_box.use_advanced_search
        elif self.force_advanced_search is not None:
            return self.force_advanced_search
        else:
            config = get_config()
            return config.setting['use_adv_search_syntax']

    def get_value_for_row_id(self, row, value):
        return row

    def setupUi(self):
        self.verticalLayout = QtWidgets.QVBoxLayout(self)
        self.verticalLayout.setObjectName('vertical_layout')
        if self.show_search:
            self.search_box = SearchBox(self, force_advanced_search=self.force_advanced_search)
            self.search_box.setObjectName('search_box')
            self.verticalLayout.addWidget(self.search_box)
        self.center_widget = QtWidgets.QWidget(self)
        self.center_widget.setObjectName('center_widget')
        self.center_layout = QtWidgets.QVBoxLayout(self.center_widget)
        self.center_layout.setObjectName('center_layout')
        self.center_layout.setContentsMargins(1, 1, 1, 1)
        self.center_widget.setLayout(self.center_layout)
        self.verticalLayout.addWidget(self.center_widget)
        self.buttonBox = QtWidgets.QDialogButtonBox(self)
        if self.show_search and self.search_type:
            self.search_browser_button = QtWidgets.QPushButton(
                _("Search in browser"), self.buttonBox)
            self.buttonBox.addButton(
                self.search_browser_button,
                QtWidgets.QDialogButtonBox.ButtonRole.ActionRole)
            self.search_browser_button.clicked.connect(self.search_browser)
        self.accept_button = QtWidgets.QPushButton(
            self.accept_button_title,
            self.buttonBox)
        self.accept_button.setEnabled(False)
        self.buttonBox.addButton(
            self.accept_button,
            QtWidgets.QDialogButtonBox.ButtonRole.AcceptRole)
        self.buttonBox.addButton(
            StandardButton(StandardButton.CANCEL),
            QtWidgets.QDialogButtonBox.ButtonRole.RejectRole)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.verticalLayout.addWidget(self.buttonBox)

    def show_progress(self):
        progress_widget = QtWidgets.QWidget(self)
        progress_widget.setObjectName('progress_widget')
        layout = QtWidgets.QVBoxLayout(progress_widget)
        text_label = QtWidgets.QLabel(_('<strong>Loading…</strong>'), progress_widget)
        text_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter | QtCore.Qt.AlignmentFlag.AlignBottom)
        gif_label = QtWidgets.QLabel(progress_widget)
        movie = QtGui.QMovie(":/images/loader.gif")
        gif_label.setMovie(movie)
        movie.start()
        gif_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter | QtCore.Qt.AlignmentFlag.AlignTop)
        layout.addWidget(text_label)
        layout.addWidget(gif_label)
        layout.setContentsMargins(1, 1, 1, 1)
        progress_widget.setLayout(layout)
        self.add_widget_to_center_layout(progress_widget)

    def show_error(self, error, show_retry_button=False):
        """Display the error string.

        Args:
            error -- Error string
            show_retry_button -- Whether to display retry button or not
        """
        error_widget = QtWidgets.QWidget(self)
        error_widget.setObjectName('error_widget')
        layout = QtWidgets.QVBoxLayout(error_widget)
        error_label = QtWidgets.QLabel(error, error_widget)
        error_label.setWordWrap(True)
        error_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        error_label.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.TextSelectableByMouse)
        layout.addWidget(error_label)
        if show_retry_button:
            retry_widget = QtWidgets.QWidget(error_widget)
            retry_layout = QtWidgets.QHBoxLayout(retry_widget)
            retry_button = QtWidgets.QPushButton(_("Retry"), error_widget)
            retry_button.clicked.connect(self.retry)
            retry_button.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Fixed))
            retry_layout.addWidget(retry_button)
            retry_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter | QtCore.Qt.AlignmentFlag.AlignTop)
            retry_widget.setLayout(retry_layout)
            layout.addWidget(retry_widget)
        error_widget.setLayout(layout)
        self.add_widget_to_center_layout(error_widget)

    def network_error(self, reply, error):
        params = {
            'url': reply.request().url().toString(QtCore.QUrl.UrlFormattingOption.RemoveUserInfo),
            'error': reply.errorString(),
            'qtcode': error,
            'statuscode': reply.attribute(
                QtNetwork.QNetworkRequest.Attribute.HttpStatusCodeAttribute)
        }
        error_msg = _("<strong>Following error occurred while fetching results:<br><br></strong>"
                      "Network request error for %(url)s:<br>%(error)s (QT code %(qtcode)d, HTTP code %(statuscode)r)<br>") % params
        self.show_error(error_msg, show_retry_button=True)

    def no_results_found(self):
        error_msg = _("<strong>No results found. Please try a different search query.</strong>")
        self.show_error(error_msg)

    def search_browser(self):
        self.tagger.search(self.search_box.query, self.search_type,
                           adv=self.use_advanced_search, force_browser=True)

    @restore_method
    def restore_state(self):
        super().restore_state()
        if self.show_search:
            self.search_box.restore_checkbox_state()

    def search_box_text(self, text):
        if self.search_box:
            self.search_box.query = text



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2018, 2021 Philipp Wolfer
# Copyright (C) 2020-2021 Laurent Monin
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from unittest.mock import Mock

from test.picardtestcase import PicardTestCase

from picard import config
from picard.dataobj import DataObject


class DataObjectTest(PicardTestCase):

    def setUp(self):
        super().setUp()
        self.obj = DataObject('id')

    def test_set_genre_inc_params_no_genres(self):
        inc = set()
        config.setting['use_genres'] = False
        require_auth = self.obj.set_genre_inc_params(inc)
        self.assertEqual(set(), inc)
        self.assertFalse(require_auth)

    def test_set_genre_inc_params_with_genres(self):
        inc = set()
        config.setting['use_genres'] = True
        config.setting['folksonomy_tags'] = False
        config.setting['only_my_genres'] = False
        require_auth = self.obj.set_genre_inc_params(inc)
        self.assertIn('genres', inc)
        self.assertFalse(require_auth)

    def test_set_genre_inc_params_with_user_genres(self):
        inc = set()
        config.setting['use_genres'] = True
        config.setting['folksonomy_tags'] = False
        config.setting['only_my_genres'] = True
        require_auth = self.obj.set_genre_inc_params(inc)
        self.assertIn('user-genres', inc)
        self.assertTrue(require_auth)

    def test_set_genre_inc_params_with_tags(self):
        inc = set()
        config.setting['use_genres'] = True
        config.setting['folksonomy_tags'] = True
        config.setting['only_my_genres'] = False
        require_auth = self.obj.set_genre_inc_params(inc)
        self.assertIn('tags', inc)
        self.assertFalse(require_auth)

    def test_set_genre_inc_params_with_user_tags(self):
        inc = set()
        config.setting['use_genres'] = True
        config.setting['folksonomy_tags'] = True
        config.setting['only_my_genres'] = True
        require_auth = self.obj.set_genre_inc_params(inc)
        self.assertIn('user-tags', inc)
        self.assertTrue(require_auth)

    def test_add_genres(self):
        self.obj.add_genre('genre1', 2)
        self.assertEqual(self.obj.genres['genre1'], 2)
        self.obj.add_genre('genre1', 5)
        self.assertEqual(self.obj.genres['genre1'], 7)

    def test_set_genre_inc_custom_config(self):
        inc = set()
        config.setting['use_genres'] = False
        config.setting['folksonomy_tags'] = False
        config.setting['only_my_genres'] = False
        custom_config = Mock()
        custom_config.setting = {
            'use_genres': True,
            'folksonomy_tags': True,
            'only_my_genres': True,
        }
        require_auth = self.obj.set_genre_inc_params(inc, custom_config)
        self.assertIn('user-tags', inc)
        self.assertTrue(require_auth)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2022 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from test.picardtestcase import (
    PicardTestCase,
    get_test_data_path,
)

from picard.disc.whipperlog import toc_from_file


class TestTocFromFile(PicardTestCase):

    def test_toc_from_file(self):
        test_log = get_test_data_path('whipper.log')
        toc = toc_from_file(test_log)
        self.assertEqual((1, 8, 149323, 150, 25064, 43611, 60890, 83090, 100000, 115057, 135558), toc)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2019 Timur Enikeev
# Copyright (C) 2019-2022 Laurent Monin
# Copyright (C) 2019-2023 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import locale
import os

from PyQt5 import (
    QtCore,
    QtGui,
    QtWidgets,
)

from picard import log
from picard.config import get_config
from picard.const.sys import IS_MACOS
from picard.util import (
    format_time,
    icontheme,
    iter_files_from_objects,
)

from picard.ui.widgets import (
    ClickableSlider,
    ElidedLabel,
    SliderPopover,
)


try:
    from PyQt5 import QtMultimedia
except ImportError as e:
    qt_multimedia_available = False
    qt_multimedia_errmsg = e.msg
else:
    qt_multimedia_available = True
    qt_multimedia_errmsg = None


def get_logarithmic_volume(player_value):
    """Return logarithmic scale volume to set slider position"""
    logarithmic_volume = QtMultimedia.QAudio.convertVolume(
        player_value / 100.,
        QtMultimedia.QAudio.VolumeScale.LinearVolumeScale,
        QtMultimedia.QAudio.VolumeScale.LogarithmicVolumeScale)
    return QtCore.qRound(logarithmic_volume * 100)


def get_linear_volume(slider_value):
    """Return linear scale volume from slider position"""
    linear_volume = QtMultimedia.QAudio.convertVolume(
        slider_value / 100.,
        QtMultimedia.QAudio.VolumeScale.LogarithmicVolumeScale,
        QtMultimedia.QAudio.VolumeScale.LinearVolumeScale)
    return QtCore.qRound(linear_volume * 100)


def get_text_width(font, text):
    metrics = QtGui.QFontMetrics(font)
    size = metrics.size(QtCore.Qt.TextFlag.TextSingleLine, text)
    return size.width()


class Player(QtCore.QObject):
    error = QtCore.pyqtSignal(object, str)

    def __init__(self, parent):
        super().__init__(parent)
        self._player = None
        self._toolbar = None
        self._selected_objects = []
        if qt_multimedia_available:
            log.debug("Internal player: QtMultimedia available, initializing QMediaPlayer")
            player = QtMultimedia.QMediaPlayer(parent)
            player.setAudioRole(QtMultimedia.QAudio.Role.MusicRole)
            self.state_changed = player.stateChanged
            self._logarithmic_volume = get_logarithmic_volume(player.volume())
            availability = player.availability()
            if availability == QtMultimedia.QMultimedia.AvailabilityStatus.Available:
                log.debug("Internal player: available, QMediaPlayer set up")
                self._player = player
                self._player.error.connect(self._on_error)
            elif availability == QtMultimedia.QMultimedia.AvailabilityStatus.ServiceMissing:
                log.warning("Internal player: unavailable, service is missing")
            else:
                log.warning("Internal player: unavailable, status=%d", availability)
        else:
            log.warning("Internal player: unavailable, %s", qt_multimedia_errmsg)

    @property
    def available(self):
        return self._player is not None

    @property
    def toolbar(self):
        return self._toolbar

    def volume(self):
        return self._logarithmic_volume

    def playback_rate(self):
        return self._player.playbackRate()

    def create_toolbar(self):
        self._toolbar = PlayerToolbar(self.parent(), self)
        return self._toolbar

    def set_objects(self, objects):
        self._selected_objects = objects
        self._toolbar.play_action.setEnabled(bool(objects))

    def play(self):
        """Play selected tracks with an internal player"""
        self._player.stop()
        playlist = QtMultimedia.QMediaPlaylist(self)
        playlist.setPlaybackMode(QtMultimedia.QMediaPlaylist.PlaybackMode.Sequential)
        playlist.addMedia([QtMultimedia.QMediaContent(QtCore.QUrl.fromLocalFile(file.filename))
                          for file in iter_files_from_objects(self._selected_objects)])
        self._player.setPlaylist(playlist)
        self._player.play()

    def pause(self, is_paused):
        """Toggle pause of an internal player"""
        if is_paused:
            self._player.pause()
        else:
            self._player.play()

    def set_volume(self, logarithmic_volume):
        """Convert to linear scale and set"""
        self._logarithmic_volume = logarithmic_volume
        self._player.setVolume(get_linear_volume(logarithmic_volume))

    def set_position(self, position):
        self._player.setPosition(position)

    def set_playback_rate(self, playback_rate):
        player = self._player
        player.setPlaybackRate(playback_rate)
        if not IS_MACOS:
            # Playback rate changes do not affect the current media playback on
            # Linux and does work unreliable on Windows.
            # Force playback restart to have the rate change applied immediately.
            player_state = player.state()
            if player_state != QtMultimedia.QMediaPlayer.State.StoppedState:
                position = player.position()
                player.stop()
                player.setPosition(position)
                if player_state == QtMultimedia.QMediaPlayer.State.PlayingState:
                    player.play()
                elif player_state == QtMultimedia.QMediaPlayer.State.PausedState:
                    player.pause()

    def _on_error(self, error):
        if error == QtMultimedia.QMediaPlayer.Error.FormatError:
            msg = _("Internal player: The format of a media resource isn't (fully) supported")
        elif error == QtMultimedia.QMediaPlayer.Error.AccessDeniedError:
            msg = _("Internal player: There are not the appropriate permissions to play a media resource")
        elif error == QtMultimedia.QMediaPlayer.Error.ServiceMissingError:
            msg = _("Internal player: A valid playback service was not found, playback cannot proceed")
        else:
            msg = _("Internal player: error, code=%(code)d, msg=%(message)s") % {
                'code': error,
                'message': self._player.errorString(),
            }
        self.error.emit(error, msg)


class PlayerToolbar(QtWidgets.QToolBar):
    def __init__(self, parent, player):
        super().__init__(_("Player"), parent)
        self.setObjectName('player_toolbar')
        self.setAllowedAreas(QtCore.Qt.ToolBarArea.TopToolBarArea
            | QtCore.Qt.ToolBarArea.BottomToolBarArea
            | QtCore.Qt.ToolBarArea.NoToolBarArea)

        self.player = player

        self.play_action = QtWidgets.QAction(icontheme.lookup('play'), _("Play"), self)
        play_tip = _("Play selected files")
        self.play_action.setToolTip(play_tip)
        self.play_action.setStatusTip(play_tip)
        self.play_action.setEnabled(False)
        self.play_action.triggered.connect(self.play)

        self.pause_action = QtWidgets.QAction(icontheme.lookup('pause'), _("Pause"), self)
        pause_tip = _("Pause or resume current playback")
        self.pause_action.setToolTip(pause_tip)
        self.pause_action.setStatusTip(pause_tip)
        self.pause_action.setCheckable(True)
        self.pause_action.setChecked(False)
        self.pause_action.setEnabled(False)
        self.pause_action.triggered.connect(self.player.pause)
        self.player.state_changed.connect(self.pause_action.setEnabled)

        self._add_toolbar_action(self.play_action)
        self._add_toolbar_action(self.pause_action)

        self.progress_widget = PlaybackProgressSlider(self, self.player)
        self.addWidget(self.progress_widget)

        config = get_config()
        volume = config.persist['mediaplayer_volume']
        self.player.set_volume(volume)
        self.volume_button = VolumeControlButton(self, volume)
        self.volume_button.volume_changed.connect(self.player.set_volume)
        self.volume_button.setToolButtonStyle(self.toolButtonStyle())
        self.addWidget(self.volume_button)

        playback_rate = config.persist['mediaplayer_playback_rate']
        self.player.set_playback_rate(playback_rate)
        self.playback_rate_button = PlaybackRateButton(self, playback_rate)
        self.playback_rate_button.playback_rate_changed.connect(self.player.set_playback_rate)
        self.playback_rate_button.setToolButtonStyle(self.toolButtonStyle())
        self.addWidget(self.playback_rate_button)

    def _add_toolbar_action(self, action):
        self.addAction(action)
        widget = self.widgetForAction(action)
        widget.setFocusPolicy(QtCore.Qt.FocusPolicy.TabFocus)
        widget.setAttribute(QtCore.Qt.WidgetAttribute.WA_MacShowFocusRect)

    def play(self):
        self.player.play()
        self.pause_action.setChecked(False)

    def setToolButtonStyle(self, style):
        super().setToolButtonStyle(style)
        self.playback_rate_button.setToolButtonStyle(style)
        self.volume_button.setToolButtonStyle(style)

    def showEvent(self, event):
        super().showEvent(event)
        self._update_popover_position()

    def _update_popover_position(self):
        popover_position = self._get_popover_position()
        self.playback_rate_button.popover_position = popover_position
        self.volume_button.popover_position = popover_position

    def _get_popover_position(self):
        if self.isFloating():
            return 'bottom'
        pos = self.mapToParent(QtCore.QPoint(0, 0))
        half_main_window_height = self.parent().height() / 2
        if pos.y() <= half_main_window_height:
            return 'bottom'
        else:
            return 'top'


class PlaybackProgressSlider(QtWidgets.QWidget):
    def __init__(self, parent, player):
        super().__init__(parent)
        self.player = player
        self._position_update = False

        tool_font = QtWidgets.QApplication.font('QToolButton')

        self.progress_slider = ClickableSlider(self)
        self.progress_slider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.progress_slider.setEnabled(False)
        self.progress_slider.setMinimumWidth(30)
        self.progress_slider.setSingleStep(1000)
        self.progress_slider.setPageStep(3000)
        self.progress_slider.valueChanged.connect(self.on_value_changed)
        self.media_name_label = ElidedLabel(self)
        self.media_name_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.media_name_label.setFont(tool_font)

        slider_container = QtWidgets.QWidget(self)
        hbox = QtWidgets.QHBoxLayout(slider_container)
        hbox.setContentsMargins(0, 0, 0, 0)
        self.position_label = QtWidgets.QLabel("0:00", self)
        self.duration_label = QtWidgets.QLabel(format_time(0), self)
        min_duration_width = get_text_width(self.position_label.font(), "8:88")
        self.position_label.setMinimumWidth(min_duration_width)
        self.duration_label.setMinimumWidth(min_duration_width)
        self.position_label.setFont(tool_font)
        self.duration_label.setFont(tool_font)
        hbox.addWidget(self.position_label)
        hbox.addWidget(self.progress_slider)
        hbox.addWidget(self.duration_label)

        vbox = QtWidgets.QVBoxLayout(self)
        vbox.setSpacing(0)
        vbox.addWidget(slider_container)
        vbox.addWidget(self.media_name_label)

        self.player._player.durationChanged.connect(self.on_duration_changed)
        self.player._player.positionChanged.connect(self.on_position_changed)
        self.player._player.currentMediaChanged.connect(self.on_media_changed)

    def on_duration_changed(self, duration):
        self.progress_slider.setMaximum(duration)
        self.duration_label.setText(format_time(duration))

    def on_position_changed(self, position):
        self._position_update = True
        self.progress_slider.setValue(position)
        self._position_update = False
        self.position_label.setText(format_time(position, display_zero=True))

    def on_media_changed(self, media):
        if media.isNull():
            self.progress_slider.setEnabled(False)
        else:
            url = media.canonicalUrl().toString()
            self.media_name_label.setText(os.path.basename(url))
            self.progress_slider.setEnabled(True)

    def on_value_changed(self, value):
        if not self._position_update:  # Avoid circular events
            self.player.set_position(value)


class PlaybackRateButton(QtWidgets.QToolButton):
    playback_rate_changed = QtCore.pyqtSignal(float)

    multiplier = 10.0

    def __init__(self, parent, playback_rate):
        super().__init__(parent)
        self.popover_position = 'bottom'
        self.rate_fmt = N_("%1.1f ×")
        button_margin = self.style().pixelMetric(QtWidgets.QStyle.PixelMetric.PM_ButtonMargin)
        min_width = get_text_width(self.font(), _(self.rate_fmt) % 8.8)
        self.setMinimumWidth(min_width + (2 * button_margin) + 2)
        self.set_playback_rate(playback_rate)
        self.clicked.connect(self.show_popover)
        tooltip = _("Change playback speed")
        self.setToolTip(tooltip)
        self.setStatusTip(tooltip)

    def show_popover(self):
        slider_value = self.playback_rate * self.multiplier
        popover = SliderPopover(
            self, self.popover_position, _("Playback speed"), slider_value)
        # In 0.1 steps from 0.5 to 1.5
        popover.slider.setMinimum(5)
        popover.slider.setMaximum(15)
        popover.slider.setSingleStep(1)
        popover.slider.setPageStep(1)
        popover.slider.setTickInterval(1)
        popover.slider.setTickPosition(QtWidgets.QSlider.TickPosition.TicksBelow)
        popover.value_changed.connect(self.on_slider_value_changed)
        popover.show()

    def on_slider_value_changed(self, value):
        playback_rate = value / self.multiplier
        self.set_playback_rate(playback_rate)
        self.playback_rate_changed.emit(self.playback_rate)

    def set_playback_rate(self, playback_rate):
        self.playback_rate = playback_rate
        label = locale.format_string(_(self.rate_fmt), playback_rate)
        self.setText(label)

    def wheelEvent(self, event):
        delta = event.angleDelta().y()
        # Incrementing repeatedly in 0.1 steps would cause floating point
        # rounding issues. Do the calculation in whole numbers to prevent this.
        new_rate = int(self.playback_rate * 10)
        if delta > 0:
            new_rate += 1
        elif delta < 0:
            new_rate -= 1
        new_rate = min(max(new_rate, 5), 15) / 10.0
        if new_rate != self.playback_rate:
            self.set_playback_rate(new_rate)
            self.playback_rate_changed.emit(new_rate)


class VolumeControlButton(QtWidgets.QToolButton):
    volume_changed = QtCore.pyqtSignal(int)

    def __init__(self, parent, volume):
        super().__init__(parent)
        self.popover_position = 'bottom'
        self.step = 3
        self.volume_fmt = N_("%d%%")
        self.set_volume(volume)
        button_margin = self.style().pixelMetric(QtWidgets.QStyle.PixelMetric.PM_ButtonMargin)
        min_width = get_text_width(self.font(), _(self.volume_fmt) % 888)
        self.setMinimumWidth(min_width + (2 * button_margin) + 2)
        self.clicked.connect(self.show_popover)
        tooltip = _("Change audio volume")
        self.setToolTip(tooltip)
        self.setStatusTip(tooltip)

    def show_popover(self):
        popover = SliderPopover(
            self, self.popover_position, _("Audio volume"), self.volume)
        popover.slider.setMinimum(0)
        popover.slider.setMaximum(100)
        popover.slider.setPageStep(self.step)
        popover.value_changed.connect(self.on_slider_value_changed)
        popover.show()

    def on_slider_value_changed(self, value):
        self.set_volume(value)
        self.volume_changed.emit(self.volume)

    def set_volume(self, volume):
        self.volume = volume
        label = _(self.volume_fmt) % volume
        self.setText(label)
        self.update_icon()

    def update_icon(self):
        if self.volume == 0:
            icon = 'speaker-0'
        elif self.volume <= 33:
            icon = 'speaker-33'
        elif self.volume <= 66:
            icon = 'speaker-66'
        else:
            icon = 'speaker-100'
        icon = icontheme.lookup(icon)
        self.setIcon(icon)

    def wheelEvent(self, event):
        delta = event.angleDelta().y()
        volume = self.volume
        if delta > 0:
            volume += self.step
        elif delta < 0:
            volume -= self.step
        volume = min(max(volume, 0), 100)
        if volume != self.volume:
            self.set_volume(volume)
            self.volume_changed.emit(volume)
